"""Module iosearch of viperleed.calc.files.

Functions for reading, processing and writing files relevant
to the search.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-19'
__license__ = 'GPLv3+'

import copy
import logging
import os
from pathlib import Path
import random
import shutil
import time

import fortranformat as ff
import numpy as np

from viperleed.calc.files import poscar
from viperleed.calc.files.beams import writeAUXEXPBEAMS
from viperleed.calc.files.iorfactor import largest_nr_grid_points
from viperleed.calc.files.iorfactor import prepare_rfactor_energy_ranges
from viperleed.calc.files.vibrocc import writeVIBROCC
from viperleed.calc.lib import leedbase
from viperleed.calc.lib.base import BackwardsReader, readIntLine
from viperleed.calc.lib.version import Version

logger = logging.getLogger(__name__)


class SearchIORaceConditionError(Exception):
    """Raised if reading of control.chem does not return the expected number
    of lines"""
    def __init__(self, message):
        super().__init__(message)


class SearchIOEmptyFileError(Exception):
    """Raised if file read for the search has no content"""
    def __init__(self, message):
        super().__init__(message)


def readSDTL_next(filename="SD.TL", offset=0):
    """
    Reads SDTL from offset to end, returns new offset and the content in
    between as string.

    Parameters
    ----------
    filename : str, optional
        The file to read from. The default is "SD.TL".
    offset : int, optional
        Where in the file to start. The default is 0.

    Returns
    -------
    offset : int
        The position to which SD.TL was read; this should be passed as offset
        when this function is called next.
    content : str
        The content of the file found between the old offset and the new.

    """
    try:
        with open(filename, "r") as rf:
            rf.seek(offset)
            content = rf.read()
            newoffset = rf.tell()
        return(newoffset, content)
    except Exception:
        logger.error("Error reading SD.TL file")
        return (offset, "")     # return old offset, no content


def readSDTL_blocks(content, whichR=0, print_info=False, n_expect=0):
    """
    Attempts to interpret a given string as one or more blocks of an SD.TL
    file.

    Parameters
    ----------
    content : str
        A block of data as read from SD.TL.
    whichR : int, optional
        Which r-factor values to use (average / integer / fractional)
    print_info : bool, optional
        Whether some basic information should be printed to logger.info and
        logger.warning. The default is False.
    n_expect : int, optional
        Number of configurations expected per block. If a block contains
        fewer, it will be ignored.

    Returns
    -------
    returnList : list
        List consists of one entry per data block. Each entry of the list is
        a tuple (gen, rfacs, configs), where gen is the generation index,
        rfacs an ordered list of R-factors, and configs the corresponding
        parameter configurations. config is a tuple listing (percent, dc)
        for each domain, with dc the parameter values of that domain.

    """
    returnList = []
    blocklist = content.split("CCCCCCCCCCC    GENERATION")[1:]
    for block in blocklist:
        gen = 0
        try:
            gen = int(block[:13])
            if print_info:
                logger.info("Reading search results from SD.TL, "
                            "generation {}".format(gen))
        except ValueError:
            if print_info:
                logger.warning("While reading SD.TL, could not interpret "
                            "generation number "+block[:13])
        lines = block.split("\n")
        rfacs = []
        configs = []
        dpars = []  # list of (list of (percent, parameters)) by domain
        for line in lines:
            if "|" in line and line[:3] != "IND":
                # this is a line with data in it
                if line.split("|")[0].strip():
                    # first line - contains R-factor
                    if dpars:
                        configs.append(tuple(dpars))
                        dpars = []
                    try:
                        rav = float(line.split("|")[2 + whichR])  # average R
                        rfacs.append(rav)
                    except ValueError:
                        if print_info:
                            logger.error("Could not read R-factor in SD.TL line:\n"
                                        + line)
                try:
                    percent = int(line.split("|")[-2].strip()[:-1])
                    valstring = line.split("|")[-1].rstrip()
                    pars = readIntLine(valstring, width=4)
                    dpars.append((percent, pars))
                except ValueError:
                    if print_info:
                        logger.error("Could not read values in SD.TL line:\n"
                                    + line)
        if dpars:
            configs.append(tuple(dpars))
        if not all([len(dp) == len(configs[0]) for dp in configs]):
            if print_info:
                logger.warning("A line in SD.TL contains fewer values than "
                            "the others. Skipping SD.TL block.")
            continue
        if gen != 0 and len(rfacs) > 0 and len(configs) > 0:
            returnList.append((gen, rfacs, tuple(configs)))
        elif len(configs) < n_expect:
            if print_info:
                logger.warning("A block in SD.TL contains fewer configurations "
                            "than expected.")
        else:
            if print_info:
                logger.warning("A block in SD.TL was read but not understood.")
    return returnList


def repeat_fetch_SDTL_last_block(which_beams,
                                 expected_params,
                                 final,
                                 max_repeats=2000,
                                 wait_time=5):
    print_info = final
    content = None
    for repeat in range(max_repeats):
        content = _fetch_SDTL_last_block(which_beams,
                                         expected_params,
                                         print_info)
        try:  # try again if content is not complete
            n_search_params_found = len(content[0][2])
        except IndexError:
            continue
        if n_search_params_found == expected_params:
            logger.debug("Read complete block from SD.TL file after "
                         f"{repeat+1} read attempt(s).")
            return content
        elif n_search_params_found >= expected_params:
            raise RuntimeError(f"Expected a maximum of {expected_params} "
                               f"parameters in SD.TL, but found "
                               "{n_search_params_found}.")
        # wait and try again
        time.sleep(wait_time/1000)
        print_info = False
    if not content:
        raise SearchIOEmptyFileError("No data found in SD.TL file.")
    # if we get here, we have exceeded the maximum number of repeats
    raise SearchIORaceConditionError(f"Could not read complete block from "
                                     "SD.TL file.")

def _fetch_SDTL_last_block(which_beams, n_expect, print_info=False):
    try:
        lines = readSDTL_end(n_expect=n_expect)
    except FileNotFoundError:
        logger.error("Could not process Search results: SD.TL file not "
                     "found.")
        raise
    except Exception:
        logger.error("Failed to get last block from SD.TL file.")
        raise
    # check if the last block contains the expected number of lines
    lines_str = "\n".join(lines)
    sdtl_content = readSDTL_blocks("\n".join(lines),
                                   whichR=which_beams,
                                   print_info=print_info,
                                   n_expect=n_expect)
    return sdtl_content


def readSDTL_end(filename="SD.TL", n_expect=0):
    """
    Reads the last generation block from the SD.TL file, starting from the
    last line containing a GENERATION label.

    Parameters
    ----------
    filename : str, optional
        Which file to read
    n_expect : int, optional
        Number of configurations expected per block. If the last block contains
        fewer, it will be ignored, instead reading the second-to-last. Set to
        0 to read the last block irrespective of length.

    Returns
    -------
    lines : str
        The contents of the file from the last line containing GENERATION to
        the end.

    """
    # get the last block from SD.TL:
    bwr = BackwardsReader(filename)
    try:
        lines = [""]
        while len(bwr.data) > 0:
            while ("CCCCCCCCCCC    GENERATION" not in lines[-1]
                   and len(bwr.data) > 0):
                lines.append(bwr.readline().rstrip())
            if len([line for line in lines if "|" in line]) > n_expect:
                break
            lines = [""]
        lines.reverse()
    finally:
        bwr.close()
    return lines


def readDataChem(rp, source, cutoff=0, max_configs=0):
    """
    Reads the data from a list of data.chem files, or a single file.

    Parameters
    ----------
    rp : Rparams
        Run parameters.
    source : either a filename, or an iterable containing multiple file names.
        The files from which to read data; usually of format 'data*.chem'.
    cutoff : float
        0 to read all, else specifies the maximum R-factor to be stored.
        All structures with higher R will be discarded.
    max_configs : int
        0 to read all, else specifies the maximum number of configurations to
        read in. Will sort configurations and read in the ones with lowest
        R-factors first, discard the rest.

    Returns
    -------
    list
        Tuples (rfac, config), where config is a tuple listing (percent, dc)
        for each domain, with dc the parameter values of that domain.

    """
    def getPercent(domain_step, dp_vals):
        """Transforms from integer domain area values to percent"""
        dsteps = int(100 / domain_step) + 1
        shifted = [v-1 for v in dp_vals]
        norm = 1
        if sum(shifted) > 0:
            norm = (dsteps-1)/sum(shifted)
        # note that TensErLEED uses implicit floor int conversion here up to
        #  version 1.6, so results may be inconsistent with SD.TL.
        pl = [int(round(v*norm))*domain_step for v in shifted[:-1]]
        pl.append(100 - sum(pl))
        return pl

    if type(source) == str:
        source = [source]
    lines = set()
    for f in source:
        with open(f, "r") as rf:
            lines.update(rf.readlines()[1:])
    returnList = []
    parslen = 0
    if max_configs == 0:
        max_configs = len(lines)
    if len(lines) > max_configs:
        lines = set(sorted(list(lines))[:max_configs])
    for line in lines:
        try:
            rfac = float(line.split("|")[0].strip())
            if cutoff != 0 and rfac > cutoff:
                continue
            valstring = line.split("|")[1].rstrip()
            pars = readIntLine(valstring, width=4)
        except (ValueError, IndexError):
            logger.debug("Could not read values in data.chem line:\n"+line)
            continue
        if len(pars) != parslen:
            if parslen == 0:
                parslen = len(pars)
            else:
                continue     # incomplete line, read while file was written
        if not rp.domainParams:
            returnList.append((rfac, ((100, pars[:-1]),)))
        else:
            dp_vals = pars[-len(rp.domainParams):]
            percent = getPercent(rp.DOMAIN_STEP, dp_vals)
            dpars = []
            for v in [len(dp.rp.searchpars) for dp in rp.domainParams]:
                dpars.append(tuple(pars[:v]))
                pars = pars[v:]
            returnList.append((rfac, tuple(zip(percent, dpars))))
    return returnList


def writeRfInfo(sl, rp, file_path="rf.info"):
    """
    Generates r-factor parameters for the search, combines them with the
    experimental beams in AUXEXPBEAMS format to make the entire input for the
    search, returns that as a string.

    Parameters
    ----------
    sl : Slab
        The Slab object used in the search. Only used for determining
        equivalent beams.
    rp : Rparams
        Run parameters.
    file_path : pathlike or str
        Pathlike to or name of the output file. If str uses
        rp.workdir /filename. The default is "rf.info".

    Returns
    -------
    output : str
        Content of the output file.
    """
    _, theo_range, _, vincr = prepare_rfactor_energy_ranges(rp)

    # find correspondence experimental to theoretical beams:
    beamcorr = leedbase.getBeamCorrespondence(sl, rp)
    # integer & fractional beams
    iorf = []
    for beam in rp.expbeams:
        if beam.hkfrac[0].denominator != 1 or beam.hkfrac[1].denominator != 1:
            iorf.append(2)
        else:
            iorf.append(1)

    if rp.TL_VERSION < Version('1.7.0'):
        formatter = {'energies': ff.FortranRecordWriter('F7.2'),
                     'int': ff.FortranRecordWriter('I3'),
                     'beams': ff.FortranRecordWriter("25I3"),
                     'weights': ff.FortranRecordWriter("25F3.1")
                     }
    else:
        formatter = {'energies': ff.FortranRecordWriter('F7.2'),
                     'int': ff.FortranRecordWriter('I4'),
                     'beams': ff.FortranRecordWriter("25I4"),
                     'weights': ff.FortranRecordWriter("25F4.1")
                     }
    output = (formatter['energies'].write([theo_range.min]).ljust(16)
              + "EMIN\n")
    output += (
        formatter['energies'].write([theo_range.max + 0.1 * vincr]).ljust(16)
        + "EMAX\n"
        )
    # !!! BULLSHIT RESULTS WHEN EMAX > MAX ENERGY IN DELTA FILES                # TODO: Issue #138
    output += (formatter['energies'].write([vincr]).ljust(16) + "EINCR\n")
    output += (formatter['int'].write([0]).ljust(16)
               + "IPR - determines amount of output to stdout\n")
    output += (formatter['energies'].write([rp.V0_IMAG]).ljust(16) + "VI\n")
    output += (formatter['energies'].write([0.]).ljust(16) + "V0RR\n")
    output += (formatter['energies'].write([rp.IV_SHIFT_RANGE.start]).ljust(16)
               + "V01\n")
    output += (formatter['energies'].write([rp.IV_SHIFT_RANGE.stop]).ljust(16)
               + "V02\n")
    output += (formatter['energies'].write([vincr]).ljust(16) + "VINCR\n")
    output += (formatter['int'].write([rp.R_FACTOR_SMOOTH]).ljust(16)
               + "ISMOTH\n")
    output += (formatter['int'].write([0]).ljust(16)
               + "EOT - 0: exp format, 1: van Hove format\n")
    output += ((formatter['int'].write([len(rp.ivbeams)])
                + formatter['int'].write([len(rp.expbeams)])).ljust(16)
               + "NTH NEX\n")
    # numbers of theoretical beams, as they correspond to experimental beams
    output += (formatter['beams'].write([n+1 for n in beamcorr]) + "\n")
    output += " DATA MITTEL (integer & fractional beams) :\n"
    output += formatter['beams'].write(iorf) + "\n"
    output += " exp - th relationship IBP, beam weights WB\n"
    output += formatter['beams'].write([n + 1 for n
                                        in range(0, len(rp.expbeams))]) + "\n"
    output += formatter['weights'].write([1]*len(rp.expbeams))+"\n"
    auxexpbeams = writeAUXEXPBEAMS(rp.expbeams, header=rp.systemName,
                                   write=True, numbers=True,
                                   version=rp.TL_VERSION)
    output += auxexpbeams

    if isinstance(file_path, str):
        _file_path = rp.workdir / file_path
    else:
        _file_path = file_path
    try:
        with open(_file_path, 'w') as wf:
            wf.write(output)
    except Exception:
        logger.error(f"Failed to write {_file_path}")
        raise
    logger.debug(f"Wrote to {_file_path} successfully")
    return output


def generateSearchInput(sl, rp, steuOnly=False, cull=False, info=True):
    """
    Generates a PARAM and a search.steu file for the search.

    Parameters
    ----------
    sl : Slab
        The Slab object used in the search. Only used if there are no domains.
    rp : Rparams
        Run parameters.
    steuOnly : bool, optional
        If True, PARAM and restrict.f will not be written. The default is
        False.
    cull : bool, optional
        If True, part of the population (given by rp.SEARCH_CULL) will be
        removed and replaced according to rp.SEARCH_CULL. The default is False.
    info : bool, optional
        If False, no info messages will be printed.

    Returns
    -------
    None

    """
    # first generate list of SearchPar objects and figure out search parameters
    rp.generateSearchPars(sl)

    if rp.indyPars == 0:
        logger.warning("The current search includes no variations.")
        rp.indyPars = 1

    # if search population is undefined, calculate a default:
    if rp.SEARCH_POPULATION == 0:
        spop = min(48, 15 + rp.indyPars)
        rp.SEARCH_POPULATION = int(np.ceil(spop / rp.N_CORES)) * rp.N_CORES

    # calculate some more things for later
    expEnergies = set([k for b in rp.expbeams for k in b.intens])
    if info:
        totalrange = rp.total_energy_range()
        logger.info("Total energy range from experimental beams is "
                    "{:.2g} eV ({} independent fit parameters, {:.2g} eV per "
                    "parameter)"
                    .format(totalrange, rp.indyPars, totalrange / rp.indyPars))
    n_theo_energies = rp.THEO_ENERGIES.n_energies
    if n_theo_energies >= len(expEnergies):
        logger.warning("Theoretical beams have more data points than "
                       "experimental beams")
        rp.setHaltingLevel(1)
        # TODO: will this crash? If so, raise here

    # merge offsets with displacement lists
    if rp.domainParams:
        attodo = [at for dp in rp.domainParams for at in dp.rp.search_atlist]
        ndom = len(rp.domainParams)
        astep = rp.DOMAIN_STEP
        nplaces = max([len(dp.rp.search_atlist) for dp in rp.domainParams])
    else:
        attodo = rp.search_atlist
        ndom = 1
        astep = 100
        nplaces = len(rp.search_atlist)
    for at in attodo:
        if at.oriState is None:
            for el in at.offset_occ:
                if el not in at.disp_occ:
                    logger.error(
                        "{} has occupation offset defined for element {}, "
                        "which does not appear in atom displacement list. "
                        "Value will be skipped, this may cause errors."
                        .format(at, el))
                else:
                    at.disp_occ[el] = [v + at.offset_occ[el]
                                       for v in at.disp_occ[el]]
                    del at.disp_occ[el] # TODO: why? we write and then delete? Should be offset_occ? If yes, loop should go over .copy
            for (disp, offset) in [(at.disp_geo, at.offset_geo),
                           (at.disp_vib, at.offset_vib),
                           (at.disp_geo, at.disp_geo_offset)]:
                dl = []
                for el in offset:
                    if el not in disp:
                        disp[el] = copy.copy(disp["all"])
                    disp[el] = [v + offset[el] for v in disp[el]]
                    dl.append(el)
                for el in dl:
                    if offset != at.disp_geo_offset:
                        del offset[el]
            at.disp_geo_offset = {"all": [np.array([0., 0., 0.])]}

    # PARAM
    output = """C Here are some global parameterization to be performed

C MNBED IS NUMBER OF EXPERIMENTAL BEAMS
C MNBTD IS NUMBER OF THEORETICAL BEAMS
C MNBMD IS MAX(MNBED,MNBTD)
      PARAMETER(MNBED = {})\n""".format(len(rp.expbeams))
    output += "      PARAMETER(MNBTD = {})\n".format(len(rp.ivbeams))
    output += "      PARAMETER(MNBMD = {})".format(max(len(rp.expbeams),
                                                       len(rp.ivbeams)))
    max_nr_data_points = largest_nr_grid_points(rp, rp.theobeams['refcalc'],
                                                False)
    output += """
C MNDATA IS MAX. NUMBER OF DATA POINTS IN EXPERIMENTAL BEAMS
      PARAMETER(MNDATA = {})""".format(max_nr_data_points)
    # array size parameter only - add some buffer -> *1.1
    output += """
C MNDATT IS NUMBER OF THEORETICAL DATA POINTS IN EACH BEAM
      PARAMETER(MNDATT = {})""".format(max_nr_data_points)
    output += """
C MPS IS POPULATION SIZE (number of independent trial structures)
      PARAMETER(MPS = {})""".format(rp.SEARCH_POPULATION)
    output += """
C MNDOM is number of domains to be incoherently averaged
      parameter (MNDOM = {})""".format(ndom)
    output += """
C MNPLACES IS NUMBER OF DIFFERENT ATOMIC SITES IN CURRENT VARIATION
      PARAMETER(MNPLACES = {})""".format(nplaces)
    output += """
C MNFILES IS MAXIMUM NUMBER OF TLEED-FILES PER SITE
      PARAMETER(MNFILES = {})""".format(rp.search_maxfiles)
    output += """
C MNCONCS IS MAX NUMBER OF CONCENTRATION STEPS PER SITE
      PARAMETER(MNCONCS = {})\n""".format(rp.search_maxconc)
    output += ("C MNPRMK IS NUMBER OF PARAMETERS - INCL ONE CONC PARAMETER "
               "PER SITE - incl. 1 per domain!\n")
    output += "      PARAMETER(MNPRMK = {})".format(len(rp.searchpars))
    output += """
C MNCSTEP IS MAX NUMBER OF VARIATIONS (geo. times therm.) IN 1 FILE
C MNATOMS IS RELICT FROM OLDER VERSIONS
      PARAMETER (MNCSTEP = {}, MNATOMS = 1)""".format(rp.mncstep)
    # write PARAM
    if not steuOnly:
        try:
            with open("PARAM", "w") as wf:
                wf.write(output)
        except Exception:
            logger.error("Failed at writing PARAM file for search.")
            raise

    # now search.steu
    output = ""
    # while producing search.steu, collect information on
    #   parameters for searchpars.info
    info = ("#".rjust(4) + "label".rjust(7) + "atom".rjust(7) + "elem".rjust(7)
            + "mode".rjust(7) + "steps".rjust(7) + "constr".rjust(7) + "\n")

    nsteps = []     # keep track of number of steps per parameter for init
    parcount = 0
    # i3 = ff.FortranRecordWriter("I3")
    i1 = ff.FortranRecordWriter("I1")
    maxgen = rp.SEARCH_MAX_GEN
    if rp.TL_VERSION < Version('1.7.0'):
        formatter = {'int': ff.FortranRecordWriter('I3'),
                     'gens': ff.FortranRecordWriter('I6'),
                     'rmut': ff.FortranRecordWriter('F7.4'),
                     'ctrl': ff.FortranRecordWriter('I3'),
                     }
        if maxgen > 999999:
            maxgen = 999999
        ctrl_width = 3
    else:
        formatter = {'int': ff.FortranRecordWriter('I5'),
                     'gens': ff.FortranRecordWriter('I7'),
                     'rmut': ff.FortranRecordWriter('F7.4'),
                     'ctrl': ff.FortranRecordWriter('I4'),
                     }
        if maxgen > 9999999:
            maxgen = 9999999
        ctrl_width = 4
    output += (formatter['int'].write([rp.indyPars]).ljust(16)
               + "number of independent parameters\n")
    f74 = ff.FortranRecordWriter('F7.4')
    output += (formatter['rmut'].write([rp.GAUSSIAN_WIDTH]).ljust(16)
               + "gaussian width control parameter RMUT\n")
    output += (formatter['int'].write([0]).ljust(16) + "initialisation for "
               "random number generator - 0: use system time, 1,...: use"
               " init\n")
    output += (formatter['int'].write([rp.R_FACTOR_TYPE]).ljust(16)
               + "1: use RPe -- 2: use R2\n")
    output += (formatter['int'].write([rp.SEARCH_BEAMS]).ljust(16)
               + "Optimization of which beam group do you want? "
               "(0=Aver,1=Int,2=Half)\n")
    if rp.TL_VERSION >= Version('1.7.1'):
        outdata = 0
        if rp.PARABOLA_FIT["type"] != "none":
            outdata = 1
        output += (formatter['int'].write([outdata]).ljust(16)
                   + "Store configurations and write data.chem files for "
                   "parabola fit (0=false)\n")
    output += formatter['gens'].write([rp.output_interval]).ljust(16) + "output interval\n"
    output += (formatter['gens'].write([maxgen]).ljust(16)
               + "desired number of generations to be performed\n")
    output += (formatter['int'].write([astep]).ljust(16)
               + "area fraction step width (%)\n")
    output += ("SD.TL           name of search document file "
               "(max. 10 characters)\n")
    output += (formatter['int'].write([ndom]).ljust(16)
               + "Number of domains under consideration\n")
    uniquenames = []
    for k in range(0, ndom):
        if ndom == 1:
            crp = rp
            csl = sl
            frompath = ""
        else:
            crp = rp.domainParams[k].rp
            csl = rp.domainParams[k].sl
            frompath = rp.domainParams[k].workdir
            info += "---- DOMAIN {} ----\n".format(k+1)
        prev_parcount = parcount
        output += ("======= Information about Domain {}: ====================="
                   "=======================\n").format(k+1)
        output += (
            formatter['int'].write([len(crp.search_atlist)]).ljust(16)
            + "Number of atomic sites in variation: Domain {}\n".format(k+1))
        displistcount = len(csl.displists)+1
        surfats = csl.getSurfaceAtoms(crp)
        for (i, at) in enumerate(crp.search_atlist):
            printvac = False
            output += (
                "------- Information about site {}: -----------------------"
                "-----------------------\n".format(i+1))
            surf = 1 if at in surfats else 0 # Flag that goes into variable NSURF used by search in GetInt
            output += (formatter['int'].write([surf]).ljust(16)
                       + "Surface (0/1)\n")
            if at.displist in csl.displists:
                dlind = csl.displists.index(at.displist) + 1
            else:
                dlind = displistcount
                displistcount += 1
            output += (formatter['int'].write([dlind]).ljust(16)
                       + "Atom number\n")
            output += (
                formatter['int'].write([len(at.known_deltas)]).ljust(16)
                + "No. of different files for Atom no. {}\n".format(i+1))
            for (j, deltafile) in enumerate(at.known_deltas):
                name = deltafile
                if frompath:  # need to get the file; if True frompath is Path
                    name = "D{}_".format(k+1) + deltafile
                    if len(name) > 15:
                        un = 1
                        while "D{}_DEL_{}".format(k+1, un) in uniquenames:
                            un += 1
                        name = "D{}_DEL_{}".format(k+1, un)
                        uniquenames.append(name)
                    try:
                        shutil.copy2(frompath / deltafile, name)
                    except Exception:
                        logger.error("Error getting Delta file {} for search"
                                     .format(os.path.relpath(
                                         frompath / deltafile)))
                        raise
                output += "****Information about file {}:\n".format(j+1)
                output += (name.ljust(16) + "Name of file {} (max. 15 "
                           "characters)\n".format(j+1))
                output += (formatter['int'].write([1]).ljust(16)
                           + "Formatted(0/1)\n")
                el = deltafile.split("_")[-2]
                if el.lower() == "vac":
                    geo = 1
                    vib = 0
                    types = 1
                    printvac = True
                else:
                    # geo0 = True
                    vib0 = True
                    for (mode, disp) in [(1, at.disp_geo), (2, at.disp_vib)]:
                        if el in disp:
                            dl = disp[el]
                        else:
                            dl = disp["all"]
                        if mode == 1:
                            geo = len(dl)
                            # if geo == 1 and np.linalg.norm(dl[0]) >= 1e-4:
                            #     geo0 = False
                        else:
                            vib = len(dl)
                            if vib == 1 and dl[0] != 0.:
                                vib0 = False
                    if vib == 1 and vib0:
                        vib = 0
                    # elif geo == 1 and geo0:
                    #     geo = 0
                    if geo > 0 and vib > 0:
                        types = 2
                    else:
                        types = 1
                output += (formatter['int'].write([types]).ljust(16)
                           + "Types of parameters in file {}\n".format(j+1))
                label = i1.write([i+1])+i1.write([j+1])
                constr = {"vib": "-", "geo": "-"}
                for mode in ["vib", "geo"]:
                    spl = [s for s in crp.searchpars if s.el == el
                           and s.atom == at and s.mode == mode]
                    if spl and spl[0].restrictTo is not None:
                        sp = spl[0]
                        if type(sp.restrictTo) == int:
                            constr[mode] = str(sp.restrictTo)
                        else:
                            constr[mode] = "#"+str(crp.searchpars.index(
                                                                sp.restrictTo)
                                                   + prev_parcount + 1)
                    elif spl and spl[0].linkedTo is not None:
                        constr[mode] = "#"+str(
                            crp.searchpars.index(spl[0].linkedTo)
                            + prev_parcount + 1)
                if vib > 0:
                    output += (formatter['int'].write([vib]).ljust(16) +
                               "vibrational steps\n")
                    parcount += 1
                    info += (str(parcount).rjust(4) + ("P"+label).rjust(7)
                             + str(at.num).rjust(7) + el.rjust(7)
                             + "vib".rjust(7) + str(vib).rjust(7)
                             + constr["vib"].rjust(7) + "\n")
                    nsteps.append(vib)
                if geo > 0:
                    output += (formatter['int'].write([geo]).ljust(16) +
                               "geometrical steps\n")
                    parcount += 1
                    info += (str(parcount).rjust(4) + ("P"+label).rjust(7)
                             + str(at.num).rjust(7) + el.rjust(7)
                             + "geo".rjust(7) + str(geo).rjust(7)
                             + constr["geo"].rjust(7) + "\n")
                    nsteps.append(geo)
            output += "****concentration steps for site no. {}\n".format(i+1)
            occsteps = len(next(iter(at.disp_occ.values())))
            output += (formatter['int'].write([occsteps]).ljust(16)
                       + "no. of concentration steps - sum must equal 1 !\n")
            for j in range(0, occsteps):
                ol = ""
                totalocc = 0
                for el in at.disp_occ:
                    ol += f74.write([at.disp_occ[el][j]])
                    totalocc += at.disp_occ[el][j]
                if printvac:
                    ol += f74.write([1 - totalocc])
                output += (ol.ljust(16)
                           + "   concentration step no. {}\n".format(j+1))
            label = i1.write([i+1])+i1.write([i+1])
            parcount += 1
            constr = "-"
            spl = [s for s in crp.searchpars if s.atom == at and
                   s.mode == "occ"]
            if spl and spl[0].restrictTo is not None:
                sp = spl[0]
                if type(sp.restrictTo) == int:
                    constr = str(sp.restrictTo)
                else:
                    constr = "#"+str(crp.searchpars.index(sp.restrictTo)+1)
            elif spl and spl[0].linkedTo is not None:
                constr = "#"+str(crp.searchpars.index(spl[0].linkedTo)+1)
            info += (str(parcount).rjust(4) + ("C"+label).rjust(7)
                     + str(at.num).rjust(7) + "-".rjust(7) + "occ".rjust(7)
                     + str(occsteps).rjust(7) + constr.rjust(7) + "\n")
            nsteps.append(occsteps)
    # add info for domain step parameters
    for k in range(0, ndom):
        parcount += 1
        if ndom == 1:
            astepnum = 1
        else:
            astepnum = int(100 / astep) + 1
        info += (str(parcount).rjust(4) + ("-".rjust(7)*3) + "dom".rjust(7)
                 + str(astepnum).rjust(7) + "-".rjust(7).rjust(7) + "\n")
        nsteps.append(astepnum)
    # now starting configuration
    output += ("Information about start configuration: (Parameters for each "
               "place, conc for each place)\n")
    if rp.SEARCH_START == "control":
        controlpath = ""
        if os.path.isfile("control.chem"):
            controlpath = "control.chem"
        elif os.path.isfile(os.path.join("SUPP", "control.chem")):
            controlpath = os.path.join("SUPP", "control.chem")
            logger.warning("No control.chem file found in working folder, "
                           "using SUPP/control.chem")
            rp.setHaltingLevel(1)
        else:
            logger.warning("No control.chem file found. Defaulting to random "
                           "starting configuration.")
            rp.SEARCH_START = "random"
            rp.setHaltingLevel(2)
    if rp.SEARCH_START == "control":
        # try reading control.chem
        # length = population +1 for header; +1 for empty line at the end
        control_chem_expected_length = rp.SEARCH_POPULATION + 2
        try:
            controllines = _read_control_chem(controlpath,
                                              control_chem_expected_length)
        except SearchIORaceConditionError:
            # Could not read last generation from control.chem
            # use backup from SD.TL instead
            if not rp.controlChemBackup:
                logger.error("Information in control.chem is incomplete "
                             "and no SD.TL data is stored. "
                             "Defaulting to random starting configuration.")
                rp.SEARCH_START = "random"
                rp.setHaltingLevel(2)
            else:
                logger.debug("Failed to read current configuration from "
                             "control.chem. "
                             "Loading last known configuraiton from stored "
                             "SD.TL data instead...")
                controllines = [s+"\n"
                                for s in rp.controlChemBackup.split("\n")]
            pass
        except OSError:
            logger.error("Error reading control.chem file. Defaulting to "
                         "random starting configuration.")
            rp.SEARCH_START = "random"
            rp.setHaltingLevel(2)

    if rp.SEARCH_START == "random":
        output += (formatter['int'].write([0]).ljust(16) +
                   "Certain start position (1) or random configuration (0)\n")
    else:
        output += (formatter['int'].write([1]).ljust(16) +
                   "Certain start position (1) or random configuration (0)\n")
        if rp.SEARCH_START == "control":
            if cull and rp.SEARCH_CULL:
                try:
                    ncull = rp.SEARCH_CULL.nr_individuals(rp.SEARCH_POPULATION)
                except ValueError:  # Too small population
                    ncull = 0
                    logger.warning(
                        "SEARCH_CULL parameter too large: would cull "
                        "entire population. Culling will be skipped."
                        )
                if any([sp.parabolaFit["min"] is not None
                        for sp in rp.searchpars]):
                    # replace one by predicted best
                    getPredicted = True
                else:
                    getPredicted = False
                nsurvive = rp.SEARCH_POPULATION - ncull
                clines = controllines[2:]
                csurvive = []
                if rp.SEARCH_CULL.type_.is_genetic or getPredicted:
                    # prepare readable clines
                    try:
                        csurvive = [readIntLine(s, width=ctrl_width)
                                    for s in clines[:nsurvive]]
                    except ValueError:
                        if rp.SEARCH_CULL.type_.is_genetic:
                            logger.warning(
                                "SEARCH_CULL: Failed to read old "
                                "configuration from control.chem, cannot run "
                                "genetic algorithm. Defaulting to cloning."
                                )
                            rp.setHaltingLevel(1)
                            csurvive = []
                for (i, line) in enumerate(clines):
                    if i < nsurvive:
                        output += line
                    elif (rp.SEARCH_CULL.type_.is_random or
                          (rp.SEARCH_CULL.type_.is_genetic and csurvive)
                          or getPredicted):
                        if getPredicted:
                            bc = None
                            if csurvive:
                                bc = csurvive[0]
                            nc = rp.getPredictConfig(
                                best_config=bc,
                                mincurv=rp.PARABOLA_FIT["mincurv"])
                            getPredicted = False
                        elif rp.SEARCH_CULL.type_.is_random:
                            nc = rp.getRandomConfig()
                        else:  # "genetic"
                            nc = rp.getOffspringConfig(csurvive)
                        ol = ""
                        for v in nc:
                            ol += formatter['ctrl'].write([v])
                        output += ol + "\n"
                    else:    # "cloning"
                        output += clines[random.randrange(0, nsurvive)]
            else:  # don't cull
                for (i, line) in enumerate(controllines):
                    if i > 1:  # first 2 lines are empty
                        output += line
        elif rp.SEARCH_START == "centered":
            for i in range(0, rp.SEARCH_POPULATION):
                ol = ""
                for n in nsteps:
                    ol += formatter['ctrl'].write([(n+1)/2])
                output += ol + "\n"
        else:    # rp.SEARCH_START == "crandom"
            pop = [rp.getCenteredConfig()]
            for i in range(1, rp.SEARCH_POPULATION):
                pop.append(rp.getRandomConfig())
            for p in pop:
                ol = ""
                for v in p:
                    ol += formatter['ctrl'].write([v])
                output += ol + "\n"
    # write search.steu
    try:
        with open("search.steu", "w") as wf:
            wf.write(output)
    except Exception:
        logger.error("Failed to write search.steu file for search.")
        raise
    # write searchpars.info
    try:
        with open("searchpars.info", "w") as wf:
            wf.write(info)
    except Exception:
        logger.error(
            "Failed to write searchpars.info file. This file is "
            "only for user information and will not affect operation.")
    if steuOnly:
        logger.debug("Wrote search.steu input for search.")
        return None
    # now restrict.f
    output = ("""
C  This subroutine serves to restrict individual parameters to certain values
C  inside the search. These values are stored in the PARIND(IPARAM,IPOP) array.
C  Parameters are counted as listed in control file search.steu, including
C  a concentration parameter after each atomic site, and the domain weight
C  parameters for each domain at the end of the PARIND array.
C
C  perform restrictions in the following fashion: e.g.
C
C     PARIND(1,IPOP) = 5
C
C  or
C
C     PARIND(5,IPOP) = PARIND(1,IPOP)
C
C  etc.

      subroutine restrict(NPRMK,NPS,PARIND,IPOP)

      INTEGER NPRMK,NPS
      INTEGER PARIND
      DIMENSION PARIND(NPRMK,NPS)

C  begin restrictions

""")
    for (i, sp) in enumerate(rp.searchpars):
        if sp.restrictTo is None:
            continue
        if type(sp.restrictTo) == int:
            output += "      PARIND({},IPOP) = {}\n".format(i+1, sp.restrictTo)
        else:
            output += "      PARIND({},IPOP) = PARIND({},IPOP)\n".format(
                i + 1, rp.searchpars.index(sp.restrictTo) + 1)
    output += ("""
C  end restrictions

      RETURN

      END
""")
    # write restrict.f
    try:
        with open("restrict.f", "w") as wf:
            wf.write(output)
    except Exception:
        logger.error("Failed to write restrict.f file for search.")
        raise
    logger.debug("Wrote input files for search: PARAM, search.steu, "
                 "restrict.f")
    return None

def _read_control_chem(control_chem_path,
                       expected_lines,
                       max_repeats=2000,
                       sleep_time=5):
    """
    Read the content of the control.chem file located at the specified path and
    return its lines.

    Parameters
    ----------
    control_chem_path : path-like
        The path to the control.chem file.
    expected_lines : int
        The expected number of lines in the control.chem file.
    max_repeats : int, optional
        The maximum number of repetitions to read the file before giving up.
        Defaults is 1000.
    sleep_time : int, optional
        The time in milliseconds to sleep between each attempt. Default is 1.

    Returns
    -------
    list of str
        The lines of the control.chem file.

    Raises
    ------
    ValueError
        If max_repeats is less than 1 or sleep_time is less than 0.
    RuntimeError
        If the number of lines in the control.chem file exceeds the
        expected_lines. This should not happen and indicates that expected_lines
        was set incorrectly.
    SearchIORaceConditionError
        If the complete control.chem file could not be read after the
        maximum number of repetitions.

    """
    if max_repeats < 1:
        raise ValueError("max_repeats must be >= 1")
    if sleep_time < 0:
        raise ValueError("sleep_time must be >= 0 ms")
    _control_chem_path = Path(control_chem_path)
    for repeat in range(max_repeats):
        with open(_control_chem_path, "r") as rf:
            control_lines = rf.readlines()
        n_control_lines = len(control_lines)
        if n_control_lines == expected_lines:
            logger.debug(f"Read complete control.chem file after {repeat+1} "
                         "read attempt(s).")
            return control_lines
        elif n_control_lines > expected_lines:
            raise RuntimeError(f"Expected at maximum {expected_lines} lines "
                               f"in control.chem, but found {n_control_lines}.")
        time.sleep(sleep_time/1000)  # in milliseconds
    raise SearchIORaceConditionError("Could not read complete "
                                     "control.chem file")

def writeSearchOutput(sl, rp, parinds=None, silent=False, suffix=""):
    """
    Modifies data in sl and rp to reflect the search result given by
    parinds, then writes POSCAR_OUT and VIBROCC_OUT.

    Parameters
    ----------
    sl : Slab
        The Slab object to be modified.
    rp : Rparams
        The run parameters object to be modified.
    parinds : iterable, optional
        The final parameter values to be applied. If None,
        rp.searchResultConfig will be used instead. Can be float, in which
        case linear interpolation between the two adjacent displacements will
        be used.
    silent : bool, optional
        Suppresses output to log. The default is False.
    suffix : str, optional
        String to be appended to the POSCAR_OUT and VIBROCC_OUT file names.

    Returns
    -------
    None

    """
    def interpolate(var, ind):
        if type(ind) == int:
            return var[ind]
        o = ind % 1
        return (o * var[int(np.ceil(ind))]
                + (1 - o) * var[int(np.floor(ind))])

    if parinds is None:
        if rp.searchResultConfig is None:
            logger.error("Failed to write search output: No configuration "
                         "passed.")
            raise RuntimeError("Failed to write search output")
        else:
            parinds = rp.searchResultConfig[0]
    # If atom and site original states are not yet saved, do it now:
    for at in sl:
        at.storeOriState()
    for site in sl.sitelist:
        if site.oriState is None:
            tmp = copy.deepcopy(site)
            site.oriState = tmp

    sl.update_cartesian_from_fractional()
    uci = np.linalg.inv(sl.ucell)
    for at in sl:
        # make list of searchpars addressing this atom:
        sps = [sp for sp in rp.searchpars if sp.atom == at and sp.el != "vac"]
        if len(sps) > 0:
            # first deal with occupation:
            par = [sp for sp in sps if sp.mode == "occ"][0]  # exactly one
            i = rp.searchpars.index(par)
            newocc = {}
            for el in at.disp_occ:
                # store as offset for now, recalculate for sites later
                at.offset_occ[el] = (interpolate(at.disp_occ[el], parinds[i]-1)
                                     - at.site.oriState.occ[el])
                newocc[el] = interpolate(at.disp_occ[el], parinds[i]-1)
                # will be used for re-calculating position
            # now vibrations:
            vibpars = [sp for sp in sps if sp.mode == "vib"]
            # can be empty, or one per element
            for par in vibpars:
                i = rp.searchpars.index(par)
                el = par.el
                if el not in at.disp_vib:
                    dict_el = "all"
                else:
                    dict_el = el
                # store as offset for now, recalculate for sites later
                at.offset_vib[el] = interpolate(at.disp_vib[dict_el],
                                                parinds[i]-1)
            # now position - recalculate right away:
            geopars = [sp for sp in sps if sp.mode == "geo"]
            # can be empty, or one per element
            if not geopars:
                continue
            totalocc = 0
            newpos = {}     # new position per element
            totalpos = np.array([0., 0., 0.])
            for par in geopars:
                i = rp.searchpars.index(par)
                el = par.el
                if el not in at.disp_geo:
                    dict_el = "all"
                else:
                    dict_el = el
                disp = np.copy(interpolate(at.disp_geo[dict_el], parinds[i]-1))
                disp[2] *= -1
                rdisp = np.dot(uci, disp)
                newpos[el] = at.oriState.pos + rdisp
                totalpos = totalpos + (newpos[el] * newocc[el])
                totalocc += newocc[el]
            if totalocc > 0:
                totalpos = totalpos / totalocc
            at.pos = totalpos
            for el in newpos:
                rel_off = newpos[el] - at.pos
                off = np.dot(sl.ucell, rel_off)
                off[2] *= -1
                at.offset_geo[el] = off
    sl.collapse_fractional_coordinates()
    sl.update_cartesian_from_fractional()
    sl.update_layer_coordinates()
    # now update site occupations and vibrations:
    for site in sl.sitelist:
        siteats = [at for at in sl if at.site == site and not at.is_bulk]
        if not siteats: # site is only found in bulk
            continue
        for el in site.occ:
            total_occ = 0
            # n_occ = 0
            total_vib = 0
            # n_vib = 0
            for at in siteats:
                if el in at.offset_occ:
                    total_occ += at.offset_occ[el]
                # n_occ += 1
                if el in at.offset_vib:
                    total_vib += at.offset_vib[el]
                # n_vib += 1
            # if n_occ > 0:
            offset_occ = total_occ/len(siteats)
            site.occ[el] = site.oriState.occ[el] + offset_occ
            # else:
            #     offset_occ = 0.0
            # if n_vib > 0:
            offset_vib = total_vib/len(siteats)
            site.vibamp[el] = site.oriState.vibamp[el] + offset_vib
            # else:
            #     offset_vib = 0.0
            for at in siteats:
                if el in at.offset_occ:
                    at.offset_occ[el] -= offset_occ
                if el in at.offset_vib:
                    at.offset_vib[el] -= offset_vib
    fn = "POSCAR_OUT" + suffix + "_" + rp.timestamp
    tmpslab = copy.deepcopy(sl)
    tmpslab.sort_original()
    try:
        poscar.write(tmpslab, filename=fn, comments="all", silent=silent)
    except OSError:
        logger.error("Exception occurred while writing POSCAR_OUT" + suffix,
                     exc_info=rp.is_debug_mode)
        rp.setHaltingLevel(2)
    if not np.isclose(rp.SYMMETRY_CELL_TRANSFORM, np.identity(2)).all():
        tmpslab = sl.make_subcell(rp, rp.SYMMETRY_CELL_TRANSFORM)
        fn = "POSCAR_OUT_mincell" + suffix + "_" + rp.timestamp
        try:
            poscar.write(tmpslab, filename=fn, silent=silent)
        except OSError:
            logger.warning(
                "Exception occurred while writing POSCAR_OUT_mincell" + suffix,
                exc_info=rp.is_debug_mode)
    fn = "VIBROCC_OUT" + suffix + "_" + rp.timestamp
    try:
        writeVIBROCC(sl, rp, filename=fn, silent=silent)
    except Exception:
        logger.error("Exception occured while writing VIBROCC_OUT" + suffix,
                     exc_info=rp.is_debug_mode)
        rp.setHaltingLevel(2)
    return



"""Section Search."""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-11'
__license__ = 'GPLv3+'

from collections import Counter
import copy
import logging
import os
from pathlib import Path
import re
import shutil
import signal
import subprocess
import sys
import time
from timeit import default_timer as timer

import numpy as np
import scipy
from sklearn.linear_model import ElasticNet
from sklearn.linear_model import Lasso
from sklearn.linear_model import LinearRegression
from sklearn.linear_model import Ridge
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import PolynomialFeatures

from viperleed.calc.classes.searchpar import SearchPar
from viperleed.calc.files import iosearch
from viperleed.calc.files import parameters
from viperleed.calc.files import searchpdf
from viperleed.calc.files.displacements import readDISPLACEMENTS_block
from viperleed.calc.lib import leedbase
from viperleed.calc.lib.checksums import validate_multiple_files
from viperleed.calc.lib.version import Version

logger = logging.getLogger(__name__)


class SearchError(Exception):
    """Base class for exceptions in this module."""


class SearchMaxIntensitiesError(SearchError):
    message = ("TensErLEED stopped due to unreasonably high beam amplitudes.\n"
               "This may be causes by convergence problems due to a too small "
               "value of LMAX. "
               "Alternatively, this error may be caused either by scatterers "
               "with very small distances as a result of a very large "
               "range used in DISPLACEMENTS."
               "Check your input files and consider increasing LMAX or "
               "decreasing the DISPLACEMENTS ranges.")

    def __init__(self, *args, **kwargs):
        super().__init__(self.message, *args, **kwargs)


class SearchSigbusError(SearchError):
    message = ("TensErLEED stopped due to a SIGBUS signal.\n"
               "This may be caused by a compiler error. "
               "If you are using gfortran, consider lowering the optimization "
               "level to '-O1' or '-O0' using the FORTRAN_COMP parameter.")

    def __init__(self, *args, **kwargs):
        super().__init__(self.message, *args, **kwargs)


class SearchInconsistentV0ImagError(SearchError):
    message = ("TensErLEED search stopped because stored Delta files were "
               "calculated with a different imaginary inner potential "
               "(optical potential) than requested in PARAMETERS. "
               "Make sure to use the same value for V0_IMAG for "
               "Delta-Amplitudes and structure search.")

    def __init__(self, *args, **kwargs):
        super().__init__(self.message, *args, **kwargs)


def processSearchResults(sl, rp, search_log_path, final=True):
    """Read the best structure from the last block of 'SD.TL' into a slab.

    Parameters
    ----------
    sl : Slab
        The Slab object to be modified
    rp : Rparams
        Run parameters.
    search_log_path : Pathlike or None
        Path to the search log. Will be checked for TensErLEED errors
        messages. If None, assumes search was not logged and skips
        checks.
    final : bool, optional
        Defines whether this is the final interpretation that *must* work and
        should go into the log, or if this will be repeated anyway. The
        default is True.

    Returns
    -------
    None.
    """
    # Read search log if available, and check for errors
    _check_search_log(search_log_path)

    # get the last block from SD.TL:
    wait_time = 5 if not final else 20
    try:
        sdtl_content = iosearch.repeat_fetch_SDTL_last_block(
            which_beams=rp.SEARCH_BEAMS,
            expected_params=rp.SEARCH_POPULATION,
            final=final, wait_time=wait_time
            )
    except iosearch.SearchIOEmptyFileError as err:
        if final:
            logger.error("No data found in SD.TL file!")
            rp.setHaltingLevel(2)
        raise RuntimeError("No data in SD.TL") from err
    except iosearch.SearchIORaceConditionError:
        return  # if we can't find a block, we also can't store it.

    (generation, rfactors, configs), *_ = sdtl_content

    assert len(rfactors) == len(configs)                                        # TODO: catch and complain. Previous version would pick only the first len(rfactors) configs, but the two should hopefully be the same length.
    writeControlChem = False                                                    # TODO: I don't think we should do this at all. We might mess with the fortran subprocess that is trying to write to the file.
    if not os.path.isfile("control.chem"):
        writeControlChem = True
    _write_control_chem(rp, generation, configs)

    # Collect populations by (r-factor, parameter indices) pairs
    populations = Counter(zip(rfactors, configs))
    if final:
        _write_info_to_log(rp, populations)


    # Pick only the best configuration (possibly for multiple
    # domains): it is the one with the highest R factor, i.e.,
    # the first one. best_doms is a list of pairs (domain_area,
    # domain_parameters) for each domain under variation
    best_doms = configs[0]
    if rp.domainParams:
        doms_info = ((dp.sl, dp.rp, dp.workdir, dp.name)
                     for dp in rp.domainParams)
    else:
        doms_info = ((sl, rp, Path.cwd(), ""),)

    # Finally write out the best structures
    home = Path.cwd()
    for (*dom_info, work, name), (_, dom_config) in zip(doms_info, best_doms):
        os.chdir(work)
        try:
            _store_and_write_best_structure(rp, *dom_info,
                                            dom_config, final)
        except Exception:                                                       # TODO: catch better
            _err = "Error while writing search output"
            if name:
                _err += f" for domain {name}"
            logger.error(_err, exc_info=rp.is_debug_mode)
            rp.setHaltingLevel(2)
            raise
        finally:
            os.chdir(home)


def _write_control_chem(rp, generation, configs):
    """Store information in configs in rp, and write it to file if needed."""
    _ctrl_chem_p = Path("control.chem")
    _write_file = not _ctrl_chem_p.is_file()
    if not _write_file:
        # check file, which generation
        try:
            with _ctrl_chem_p.open("r", encoding="utf-8") as control_chem:
                _, header, *other_lines = control_chem
        except (OSError, ValueError):
            # try overwriting, just to be safe
            _write_file = True
        else:
            try:
                gen_nr = int(header.split("No.")[-1].split(":")[0])
            except (IndexError, ValueError):
                _write_file = True
            else:
                _write_file = (gen_nr != generation
                               or len(other_lines) < rp.SEARCH_POPULATION)

    # Prepare lines to store as backup, and perhaps write to file
    lines = ["\n",  # Starts with one empty line
             f"Parameters of generation No.{generation:>6d}:\n"]
    astep = rp.DOMAIN_STEP if rp.domainParams else 100
    if rp.TL_VERSION < Version('1.7.0'):
        ctrl_width = 3
    else:
        ctrl_width = 4
    for dom_pars in configs:
        areapars = (int(percent/astep) + 1 for percent, _ in dom_pars)
        lines.append(
            "".join(f"{i:>{ctrl_width}d}" for _, par in dom_pars for i in par)
            + "".join(f"{a:>{ctrl_width}d}" for a in areapars) + "\n"
            )
    rp.controlChemBackup = "".join(lines)
    if not _write_file:
        return
    try:
        with _ctrl_chem_p.open("w", encoding="utf-8") as control_chem:
            control_chem.writelines(lines)                                      # TODO: I don't think we should do this at all. We might mess with the fortran subprocess that is trying to write to the file.
    except Exception:
        logger.error("Failed to write control.chem")
        rp.setHaltingLevel(1)


def _write_info_to_log(rp, pops):
    """Write information about populations to the current logger."""
    (most_common_r, _), n_most_common = pops.most_common(1)[0]
    if n_most_common == rp.SEARCH_POPULATION:
        logger.info("All trial structures converged to the same "
                    f"configuration (R = {most_common_r:.4f})")
        return

    (best_r, _), n_best = next(iter(pops.items()))
    # if any(v > rp.SEARCH_POPULATION / 2 for v in popcount):                   # TODO: used to be like this, but I think the current one is the same.
    if n_most_common > rp.SEARCH_POPULATION / 2:
        info = ("The search outcome is dominated by one configuration "
                f"(population {n_most_common} of {rp.SEARCH_POPULATION}, "
                f"R = {most_common_r:.4f})\n")
    else:
        info = ("The search outcome is not dominated by any "
                "configuration. The best result has population "
                f"{n_best} (of {rp.SEARCH_POPULATION}), "
                f"R = {best_r:.4f}\n")
    # if len(pops) > 1:                                                         # TODO: what follows used to be done only in this case, but I think it should already be covered by the first "if n_most_common == rp.SEARCH_POPULATION"
    info += ("The best configurations are:\nPOP       R |")
    if rp.domainParams:
        info += " area |"
    info += " PARAMETERS\n"
    for (pop_r, pop_pars), n_pops in pops.most_common(5):
        info += f"{n_pops:>3}  {pop_r:.4f} |"
        for j, (percent, pars) in enumerate(pop_pars):
            if j:
                info += "            |"
            if rp.domainParams:
                info += f" {percent:>3}% |"
            info += "".join(f"{v:>3}" for v in pars) + "\n"
    logger.info(info)


def _store_and_write_best_structure(rp, dom_slab, dom_rp, best_config, final):
    """Modify dom_slab to the parameters in best_config. Write it out.

    If there is a predicted optimum for the parameters (via parabola
    fit), POSCAR_OUT and VIBROCC_OUT files are also written for this
    predicted optimum. However, `dom_slab` will always contain the
    configuration passed via `best_config` at the end of this call.

    Parameters
    ----------
    rp : Rparams
        The PARAMETERS for the whole calculation.
    dom_slab : Slab
        The slab of the domain to be modified, and used to write
        POSCAR_OUT and VIBROCC_OUT files.
    dom_rp : Rparams
        The PARAMETERS for this specific domain to be written out.
    best_config : Sequence
        The indices of the optimal parameters to be used for this
        domain's configuration.

    Returns
    -------
    None.
    """
    # The order here matters, as the final write operation
    # will store the new states (e.g., for later searches)
    # in dom_slab, its atoms, and its sites.
    if any(sp.parabolaFit["min"] for sp in rp.searchpars):
        parab_inds = list(best_config)
        for j, sp in enumerate(dom_rp.searchpars):
            if sp.parabolaFit["min"] is not None:
                parab_inds[j] = sp.parabolaFit["min"]
        iosearch.writeSearchOutput(dom_slab, dom_rp, parab_inds,
                                   silent=True, suffix="_parabola")
    iosearch.writeSearchOutput(dom_slab, dom_rp, best_config, silent=not final)


def _check_search_log(search_log_path):
    """Check the file at `search_log_path` for fatal TensErLEED errors.

    Parameters
    ----------
    search_log_path : str or Path or None
        Path to the log file. No checking is performed if None.

    Raises
    ------
    OSError
        When failing to open or read a non-None `search_log_path`.
    RuntimeError                                                                # TODO: would be nice to raise a nicer error. Some TensErLEEDExecutionError perhaps?
        If an error message is found in the search log file that
        corresponds to a fatal TensErLEED error.
    """
    if search_log_path is None:
        return

    try:
        with open(search_log_path, "r", encoding="utf-8") as search_log:
            log_content = search_log.read()
    except OSError:
        logger.error(f"Could not read search logfile {search_log_path}. "
                     "Execution will continue, but TensErLEED errors related "
                     "to the search may not be detected properly.")
        return

    gen_err_msg = ("TensErLEED Error encountered in search. "
                   "Execution cannot proceed.")
    # (1) unreasonably high amplitude values:
    if ("MAX. INTENS. IN THEOR. BEAM" in log_content
            and "IS SUSPECT" in log_content
            and "****** STOP PROGRAM ******" in log_content):
        raise SearchMaxIntensitiesError from None
    # (2) SIGBUS signal
    elif ("received signal SIGBUS" in log_content
          and "undefined portion of a memory object" in log_content):
        raise SearchSigbusError from None
    # (3) different V0_IMAG between Deltas and Search
    elif ("Average optical potential value in rf.info is incorrect:" 
          in log_content):
        raise SearchInconsistentV0ImagError from None


def parabolaFit(rp, datafiles, r_best, x0=None, max_configs=0, **kwargs):
    """
    Performs a parabola fit to all rfactor/configuration data in r_configs.
    Stores the result in the rp searchpar objects. Keyword arguments are read
    from rp by default.

    Parameters
    ----------
    rp : Rparams
        The Rparams object containing runtime information
    datafiles : list of str
        The files to read data from
    r_best : float
        Best known R-factor so far
    x0 : numpy.array, optional
        A starting guess for the minimum
    max_configs : int
        0 to read all, else specifies the maximum number of configurations to
        read in. Will sort configurations and read in the ones with lowest
        R-factors first, discard the rest.
    Keyword Arguments:
        localize : real, optional
            If not zero, r_configs will be reduced to only use points close to
            the current best result. The value determines the region to be used
            ascompared to the full data, i.e. 0.5 would use only half of the
            space in each dimesion.
        mincurv : float, optional
            The minimum curvature that the N-dimensional paraboloid is
            required to have along the axis of a given parameter, in order for
            the minimum and curvature to be saved for that parameter.
        type : str, optional
            Defines the regression model to use in the fit. Allowed values are
            LinearRegression (default), Ridge, Lasso, and ElasticNet. Note that
            for the fit, parameters are centered around the current best
            configuration. Therefore, all models except LinearRegression put
            some penalty not only on curvature, but also on the parameter
            offset from the best-R configuration.
            Setting 'type' to 'none' turns off parabola fit entirely.
        alpha : float, optional
            The alpha value used by Ridge, Lasso and ElasticNet regression. The
            default is 1e-5.

    Returns
    -------
    np.array
        The result vector in parameter space, consisting of the coordinates of
        the parabola minimum for those dimensions where the parabola curvature
        is greater than mincurv, and the values from the configuration with
        the lowest R-factor for all dimensions. Can be passed back as x0 in
        the next iteration.
    float
        Predicted minimum R-factor.

    """

    def optimizerHelper(array, func):
        return func(array.reshape(1, -1))

    def castToMatrix(features, dim):
        """Casts the weights of PolynomialFeatures of degree 2 into a square
        symmetric 2D array, discarding linear features."""
        m = np.zeros((dim, dim))
        start = dim+1
        for i in range(0, dim):
            m[i][i:] = features[start:start + dim - i]
            start += dim - i
        m = 0.5*(m + m.T)
        return m

    if not datafiles:
        return None, None
    if "type" not in kwargs:
        which_regression = rp.PARABOLA_FIT["type"]
    else:
        which_regression = kwargs["type"]
    if which_regression == "none":
        return None, None
    for a in ("localize", "mincurv", "alpha", "type"):
        if a not in kwargs:
            kwargs[a] = rp.PARABOLA_FIT[a]

    # read data
    r_cutoff = 1.0
    if datafiles:
        varR = np.sqrt(8*abs(rp.V0_IMAG) / rp.total_energy_range())*r_best
        rc = np.array((iosearch.readDataChem(
            rp, datafiles,
            cutoff=r_best + r_cutoff * varR,
            max_configs=max_configs
            )), dtype=object)
    if len(rc) < 100*rp.indyPars:
        return None, None

    rfacs, configs = rc[:, 0].astype(float), rc[:, 1]
    localizeFactor = kwargs["localize"]
    if localizeFactor == 0:
        localizeFactor = 1  # no localization
    # reduce to independent parameters during fit
    if not rp.domainParams:
        sps = [sp for sp in rp.searchpars if sp.el != "vac" and
               sp.mode != "dom" and sp.steps*localizeFactor >= 3 and
               sp.linkedTo is None and sp.restrictTo is None]
        indep_pars = np.array([*np.array([*np.array(configs, dtype=object)],    # TODO: pylint complains about too-many-function-args. I don't understand the line
                                         dtype=object)
                               .reshape(-1, 1, 2)[:, 0, 1]])
        indep_pars = np.delete(indep_pars, [i for i in
                                            range(len(rp.searchpars)-1)
                                            if rp.searchpars[i] not in sps], 1)
    else:
        # first the percentages:
        sps = [sp for sp in rp.searchpars if sp.mode == "dom"]
        reshaped = (np.array([*np.array(configs, dtype=object)], dtype=object)  # TODO: pylint complains about too-many-function-args. I don't understand the line
                    .reshape(-1, len(rp.domainParams), 2))
        indep_pars = reshaped[:, :, 0].astype(int)  # contains the percentages
        # then the 'real' parameters:
        for (j, dp) in enumerate(rp.domainParams):
            new_sps = [sp for sp in dp.rp.searchpars if
                       sp.el != "vac" and sp.steps*localizeFactor >= 3 and
                       sp.linkedTo is None and sp.restrictTo is None and
                       sp.mode != "dom"]
            new_ip = np.array([*reshaped[:, j, 1]])
            new_ip = np.delete(new_ip, [i for i in range(len(dp.rp.searchpars))
                               if dp.rp.searchpars[i] not in new_sps], 1)
            sps.extend(new_sps)
            indep_pars = np.append(indep_pars, new_ip, axis=1)
    indep_pars = indep_pars.astype(float)
    best_config = np.copy(indep_pars[np.argmin(rfacs)])
    sps_original = sps[:]

    which_regression = kwargs["type"].lower()
    alpha = kwargs["alpha"]
    if which_regression == 'lasso':
        polyreg = make_pipeline(PolynomialFeatures(degree=2),
                                Lasso(alpha=alpha, normalize=True))             # TODO: incorrect keyword normalize
    elif which_regression == 'ridge':
        polyreg = make_pipeline(PolynomialFeatures(degree=2),
                                Ridge(alpha=alpha, normalize=True))             # TODO: incorrect keyword normalize
    elif which_regression == 'elasticnet':
        polyreg = make_pipeline(PolynomialFeatures(degree=2),
                                ElasticNet(alpha=alpha, normalize=True))        # TODO: incorrect keyword normalize
    else:
        if which_regression not in ('linearregression', 'linear'):
            logger.warning("Regression model {} not found, parabola fit "
                           "defaulting to linear regression."
                           .format(which_regression))
        which_regression = "linearregression"
        polyreg = make_pipeline(PolynomialFeatures(degree=2),
                                LinearRegression())
    xmin = np.copy(indep_pars[np.argmin(rfacs), :])
    rr = np.sqrt(8*abs(rp.V0_IMAG) / rp.total_energy_range())
    ip_tofit = np.copy(indep_pars)
    rf_tofit = np.copy(rfacs)
    # throw out high R-factors - TODO: perhaps also throw out highest X% ?
    to_del = np.where(rf_tofit > min(rf_tofit) + 3*rr)
    ip_tofit = np.delete(ip_tofit, to_del, axis=0)
    rf_tofit = np.delete(rf_tofit, to_del, axis=0)
    # center on best config; important because offset from the
    #  best configuration may also be penalized in regression
    ip_tofit = ip_tofit - xmin
    # fit
    polyreg.fit(ip_tofit, rf_tofit)
    # now find minimum within bounds (still centered around xmin)
    bounds = np.array([(1, sp.steps) for sp in sps]) - xmin.reshape((-1, 1))
    if x0 is None:  # x0 is the starting guess
        x0 = np.copy(xmin)
    # else:
    #     x0 = np.delete(x0, deletedPars)
    parab_min = scipy.optimize.minimize(
        optimizerHelper, x0, args=(polyreg.predict,), method='L-BFGS-B',
        bounds=bounds)
    predictR = parab_min.fun[0]
    # find errors
    m = castToMatrix(polyreg.named_steps[which_regression].coef_, len(sps))
    w, v = np.linalg.eig(m)
    # error along main axes
    d = np.copy(np.diag(m))
    d[d <= 0] = np.nan
    with np.errstate(invalid="ignore"):
        err_unco = 2 * np.sqrt(rr * predictR / d)
        err_unco[err_unco < 0] = np.nan
    err_unco = np.sqrt(err_unco)
    # error along eigenvectors
    w2 = np.copy(w)
    w2[w2 == 0] = 1e-100  # to avoid divide-by-zero; dealt with below
    with np.errstate(invalid="ignore"):
        err_ev = 2 * np.sqrt(rr * predictR / w2)
        err_ev[err_ev < 0] = 0  # dealt with below
    err_ev = np.sqrt(err_ev)
    # correlated error
    err_co = np.dot(v**2, err_ev)
    for i in range(len(err_co)):
        if any(w[j] <= 0 and v[i, j] >= 0.1 for j in range(len(w))):
            err_co[i] = np.nan
    # !!! TODO: maybe discard parameters and re-fit?

    # deletedPars = []
    # while True:
    #     ip_tofit = np.copy(indep_pars)
    #     rf_tofit = np.copy(rfacs)
    #     xmin = np.copy(ip_tofit[np.argmin(rfacs), :])
    #     if kwargs["localize"] != 0:
    #         # discard points that are far from global min in any dimension
    #         base = np.copy(xmin)  # parameter vector of best conf
    #         for i in range(len(base)):
    #             r = sps[i].steps*localizeFactor*0.5
    #             base[i] = max(base[i], 1 + r)
    #             base[i] = min(base[i], sps[i].steps - r)
    #         dist_norm = np.array([1/(sp.steps-1) for sp in sps])
    #         dist = abs(dist_norm*(ip_tofit - base))
    #         maxdist = np.max(dist, 1)
    #         # dist = np.linalg.norm(dist_norm*(indep_pars - base), axis=1)
    #         # cutoff = np.percentile(dist, 30)
    #         # cutoff = np.max(dist) * 0.75
    #         cutoff = 0.5*localizeFactor
    #         # weights = [1 if d <= cutoff else 0.1 for d in dist]
    #         to_del = [i for i in range(len(maxdist)) if maxdist[i] > cutoff]
    #         ip_tofit = np.delete(ip_tofit, to_del, axis=0)
    #         rf_tofit = np.delete(rf_tofit, to_del, axis=0)
    #     # center on best config; important because offset from the
    #     #  best configuration may also be penalized in regression
    #     ip_tofit = ip_tofit - xmin
    #     # check curvature of the parabolas per parameter
    #     polyreg.fit(ip_tofit, rf_tofit)
    #     curvs = [(polyreg.named_steps[which_regression]
    #               .coef_[polyreg.named_steps['polynomialfeatures']
    #               .get_feature_names().index('x{}^2'.format(i))])
    #              * sp.steps**2
    #              for i, sp in enumerate(sps)]
    #     if min(curvs) >= kwargs["mincurv"]:
    #         if len(rf_tofit) < 50*len(sps):  # too few points for good fit
    #             # logger.debug("{} points - too few for fit of {} dimensions"
    #             #               .format(len(rf_tofit), len(sps)))
    #             for sp in rp.searchpars:
    #                 sp.parabolaFit = {"curv": None, "min": None}
    #             return None, None
    #         break  # all parabola dimensions now have reasonable shape
    #     if len(sps) == 1:
    #         # logger.debug("Found no parameter with sufficient curvature")
    #         sps[0].parabolaFit = {"curv": None, "min": None}
    #         return None, None  # found no parameter with parabola shape
    #     # throw out the parameters with lowest curvature; repeat.
    #     #   discard: min. 1, max. all with low curv.
    #     k = min(max(1, int(len(curvs)*0.2)),
    #             sum(c < kwargs["mincurv"] for c in curvs))
    #     ind = np.argpartition(curvs, k)[:k]
    #     for i in ind:
    #         sps[i].parabolaFit = {"curv": None, "min": None}
    #         deletedPars.append(sps_original.index(sps[i]))
    #     sps = [sps[i] for i in range(len(sps)) if i not in ind]
    #     indep_pars = np.delete(indep_pars, ind, axis=1)
    # for (i, sp) in enumerate(sps):
    #     sp.parabolaFit["curv"] = curvs[i]

    # logger.debug("Parabola fit: {}/{} parameters were fit with {} points "
    #              "({:.4f} s)".format(len(sps), len(sps_original),
    #                                  len(rf_tofit), (timer() - starttime)))

    for (i, sp) in enumerate(sps):
        sp.parabolaFit["min"] = parab_min.x[i] + xmin[i]
        sp.parabolaFit["err_co"] = err_co[i]
        sp.parabolaFit["err_unco"] = err_unco[i]
        best_config[sps_original.index(sp)] = parab_min.x[i] + xmin[i]
    for sp in [sp for sp in rp.searchpars if isinstance(sp.linkedTo, SearchPar)
               or isinstance(sp.restrictTo, SearchPar)]:
        if isinstance(sp.linkedTo, SearchPar):
            sp.parabolaFit = copy.copy(sp.linkedTo.parabolaFit)
        elif isinstance(sp.restrictTo, SearchPar):
            sp.parabolaFit = copy.copy(sp.restrictTo.parabolaFit)
    return (best_config, predictR)


def search(sl, rp):
    """
    Generates input for and runs the search, then interprets output.

    Parameters
    ----------
    sl : Slab
        Slab object containing atom, site and layer information.
    rp : Rparams
        The run parameters.

    Raises
    ------
    RuntimeError
        Raised if execution cannot continue.

    Returns
    -------
    None.

    """

    def kill_process(proc, default_pgid=None):
        """Cleanly kill the mpirun subprocess and its children. If the process
        is not alive any more (and therefore the pgid cannot be determined),
        will instead try to terminate the default_pgid, if passed."""
        # determine pgid
        try:
            pgid = os.getpgid(proc.pid)                                         # TODO: only UNIX!
        except ProcessLookupError:
            pgid = default_pgid
        # kill main process
        try:
            proc.kill()
            proc.wait()
        except ProcessLookupError:
            pass   # already dead
        if pgid is not None:
            # kill children
            try:
                os.killpg(pgid, signal.SIGTERM)                                 # TODO: only UNIX
                if os.name == "nt":
                    os.waitpid(pgid)
                else:
                    os.waitpid(-pgid, 0)
            except (ProcessLookupError, ChildProcessError):
                pass  # already dead or no children

    rp.searchResultConfig = None
    if rp.domainParams:
        initToDo = [(dp.rp, dp.sl, dp.workdir) for dp in rp.domainParams]
    else:
        initToDo = [(rp, sl, rp.workdir)]
    for (rpt, slt, path) in initToDo:
        # read DISPLACEMENTS block
        if not rpt.disp_block_read:
            readDISPLACEMENTS_block(rpt, slt,
                                    rpt.disp_blocks[rpt.search_index])
            rpt.disp_block_read = True
        # get Deltas
        if 2 not in rpt.runHistory:
            if "Tensors" in rpt.manifest:
                logger.error("New tensors were calculated, but no new delta "
                             "files were generated. Cannot execute search.")
                raise RuntimeError("Delta calculations was not run for "
                                   "current tensors.")
            leedbase.getDeltas(rpt.TENSOR_INDEX, basedir=path,
                               targetdir=path, required=True)
    rp.updateCores()
    # generate rf.info
    try:
        rf_info_path = rp.workdir / "rf.info"
        rf_info_content = iosearch.writeRfInfo(sl, rp, file_path=rf_info_path)
    except Exception:
        logger.error("Error generating search input file rf.info")
        raise
    # generate PARAM and search.steu
    #   needs to go AFTER rf.info, as writeRfInfo may remove expbeams!
    try:
        iosearch.generateSearchInput(sl, rp)
    except Exception:
        logger.error("Error generating search input")
        raise
    # delete old control.chem if present - not earlier because can be input!
    try:
        os.remove("control.chem")
    except FileNotFoundError:
        pass
    if not rp.indyPars:  # never for calculations with domains                  # !!! CHECK
        logger.info("Found nothing to vary in search. Will proceed "
                    "directly to writing output and starting SUPERPOS.")
        this_domain_pars = [1] * len(rp.searchpars)                             # TODO: Used to be "[1] * (len) - 1" --> TypeError
        for (i, sp) in enumerate(rp.searchpars):
            if isinstance(sp.restrictTo, int):
                this_domain_pars[i] = sp.restrictTo
            elif isinstance(sp.restrictTo, SearchPar):
                this_domain_pars[i] = rp.searchpars.index(sp.restrictTo) + 1
            elif isinstance(sp.linkedTo, SearchPar):
                this_domain_pars[i] = rp.searchpars.index(sp.linkedTo) + 1
        rp.searchResultConfig = ((100, this_domain_pars),)                      # TODO: Used to assign straight to rp.searchResultConfig[i] -> IndexError as there was only a single element.
        iosearch.writeSearchOutput(sl, rp)
        return None

    if rp.SUPPRESS_EXECUTION:
        logger.warning("SUPPRESS_EXECUTION parameter is on. Search "
                       "will not proceed. Stopping...")
        rp.setHaltingLevel(3)
        return None

    # check for mpirun, decide whether to use parallelization
    usempi = True
    if rp.N_CORES == 1:
        logger.warning(
            "The N_CORES parameter is set to 1. The search will be run "
            "without multiprocessing. This will be much slower!")
        usempi = False

    if usempi and shutil.which("mpirun", os.X_OK) is None:
        usempi = False
        logger.warning(
            "mpirun is not present. Search will be compiled and executed "
            "without parallelization. This will be much slower!")

    _find_compiler = None
    if usempi and not rp.FORTRAN_COMP_MPI[0]:
        _find_compiler = rp.getFortranMpiComp
    elif not rp.FORTRAN_COMP[0]:
        _find_compiler = rp.getFortranComp

    if _find_compiler:
        try:
            _find_compiler()
        except Exception as exc:
            _mpi = 'mpi ' if usempi else ''
            logger.error(f"No fortran {_mpi}compiler found, cancelling...")
            raise FileNotFoundError("Fortran compile error") from exc
    # get fortran files
    try:
        tl_source = rp.get_tenserleed_directory()
        tl_path = tl_source.path
        srcpath = tl_path / 'src'
        if usempi:
            src_file = next(srcpath.glob('search.mpi*'), None)
        else:
            src_files = (f for f in srcpath.glob('search*')
                         if 'mpi' not in f.name)
            src_file = next(src_files, None)
        if src_file is None:
            raise FileNotFoundError(f"No Fortran source for search in {tl_path}")
        shutil.copy2(src_file, src_file.name)
        libpath = tl_path / 'lib'
        libpattern = "lib.search"
        if usempi and rp.TL_VERSION <= Version('1.7.3'):
            libpattern += ".mpi"
        lib_file = next(libpath.glob(libpattern + "*"), None)
        if lib_file is None:
            raise FileNotFoundError(f"File {libpattern}.f not found.")

        # copy to work dir
        shutil.copy2(lib_file, lib_file.name)
        hashing_file = next(libpath.glob("intarr_hashing*.f90"), None)
        if hashing_file:
            hashname = hashing_file.name
            shutil.copy2(hashing_file, hashname)
        else:
            hashname = ""

        randname = "MPIrandom_.o" if usempi else "random_.o"
        if rp.TL_VERSION <= Version('1.7.3'):
            # these are short C scripts - use pre-compiled versions

            # try to copy randomizer lib object file
            try:
                shutil.copy2(libpath / randname, randname)
            except FileNotFoundError:
                logger.error("Could not find required random_.o object file. "
                             "You may have forgotten to compile random_.c.")
                raise

        globalname = "GLOBAL"
        shutil.copy2(srcpath / globalname, globalname)
    except Exception:
        logger.error("Error getting TensErLEED files for search: ")
        raise

    # Validate TensErLEED input files
    if not rp.TL_IGNORE_CHECKSUM:
        files_to_check = (lib_file,
                          src_file,
                          srcpath / globalname,
                          hashing_file)
        validate_multiple_files(files_to_check, logger,
                                "search", rp.TL_VERSION)

    # compile fortran files
    searchname = f"search-{rp.timestamp}"
    if usempi:
        fcomp = rp.FORTRAN_COMP_MPI
    else:
        fcomp = rp.FORTRAN_COMP
    logger.info("Compiling fortran input files...")
    # compile
    # compile task could be inherited from general CompileTask (issue #43)
    ctasks = [(f"{fcomp[0]} -o lib.search.o -c", lib_file.name, fcomp[1])]
    if hashname:
        ctasks.append((f"{fcomp[0]} -c", hashname, fcomp[1]))
    ctasks.append((f"{fcomp[0]} -o restrict.o -c", "restrict.f", fcomp[1]))
    format_tag = ""
    _fixed_format = any(f.endswith('.f90')
                        for f in (lib_file.name, src_file.name, hashname))
    if _fixed_format:
        format_tag = "-fixed"
        if any(s in fcomp[0] for s in ("gfortran", "mpifort")):
            # assume that mpifort also uses gfortran
            format_tag = "--fixed-form"  # different formatting string
    ctasks.append(
        (f"{fcomp[0]} -o search.o -c {format_tag}", src_file.name, fcomp[1])
        )
    to_link = "search.o lib.search.o restrict.o"
    if rp.TL_VERSION <= Version('1.7.3'):
        to_link += f" {randname}"
    if hashname:
        to_link += " intarr_hashing.o"
    ctasks.append((f"{fcomp[0]} -o {searchname}", to_link, fcomp[1]))
    compile_log = "compile-search.log"
    try:
        leedbase.fortran_compile_batch(ctasks, logname=compile_log)
    except Exception:
        leedbase.copy_compile_log(rp, Path(compile_log),
                                  log_name="search-compile")
        logger.error("Error compiling fortran files: ", exc_info=True)
        raise
    logger.debug("Compiled fortran files successfully")
    # run
    if rp.LOG_SEARCH:
        search_log_path = (rp.workdir / searchname).with_suffix(".log")
        logger.info(f"Search log will be written to file {search_log_path}.")
    else:
        search_log_path = None
    # if there is an old SD.TL file, it needs to be removed
    if os.path.isfile("SD.TL"):
        try:
            os.remove("SD.TL")
        except OSError:
            logger.warning("Failed to delete old SD.TL file. This may "
                           "cause errors in the interpretation of search "
                           "progress.")
    # same for old data.chem
    _datachem_re = re.compile(r'data\d+\.chem$', re.IGNORECASE)
    for file in Path().glob("*"):
        if not _datachem_re.match(file.name):
            continue
        try:
            file.unlink()
        except OSError:
            logger.warning(f"Failed to delete old {file} file. This "
                           "may cause errors in the parabola fit.")
    # get config size
    config_size = (
        sys.getsizeof((1., tuple()))
        + sys.getsizeof((tuple(),) * max(1, len(rp.domainParams)))
        + sys.getsizeof((1, tuple()))
        )
    if rp.domainParams:
        config_size += sum(
            sys.getsizeof((0,) * len(dp.rp.searchpars))
            + sys.getsizeof(1) * len(dp.rp.searchpars)
            for dp in rp.domainParams
            )
    else:
        config_size += (sys.getsizeof((0,) * len(rp.searchpars))
                        + sys.getsizeof(1) * len(rp.searchpars))
    max_memory_parab = 1e9   # maximum memory in bytes to use on parabola fit
    max_read_configs = int(max_memory_parab / config_size)
    # start search process
    repeat = True
    genOffset = 0
    last_debug_write_gen = 0
    gens = []  # generation numbers in SD.TL, but continuous if search restarts
    markers = []
    rfaclist = []
    parab_x0 = None     # starting guess for parabola                           # TODO: would be nicer to incorporate it into a ParabolaFit class
    rfac_predict = []  # tuples (gen, r) from parabola fit                      # TODO: would be nicer to incorporate it into a ParabolaFit class
    realLastConfig = {"all": [], "best": [], "dec": []}                         # TODO: would be nicer with a SearchConfigurationTracker class that has an .update
    realLastConfigGen = {"all": 0, "best": 0, "dec": 0}
    convergedConfig = {"all": None, "best": None, "dec": None}
    lastconfig = None
    rp.searchMaxGenInit = rp.SEARCH_MAX_GEN
    absstarttime = timer()
    tried_repeat = False        # if SD.TL is not written, try restarting
    pgid = None
    logger.info("Starting search. See files Search-progress.pdf "
                "and SD.TL for progress information.")
    while repeat:                                                               # TODO: all this mess would be nicer to handle with a state machine approach. This would at least help readability on the various ways things are handled
        repeat = False
        interrupted = False
        proc = None
        if usempi:
            command = ["mpirun", "-n", str(rp.N_CORES),
                       os.path.join(".", searchname)]
        else:
            command = [os.path.join('.', searchname),]
        # if LOG_SEARCH -> log search
        if search_log_path:
            log_exists = search_log_path.is_file()
            search_log_f = search_log_path.open("a")
            if log_exists:  # log file existed before
                search_log_f.write("\n\n-------\nRESTARTING\n-------\n\n")
        else:
            search_log_f = subprocess.DEVNULL
        # NB: log file may be open and must be closed!
        # Create search process
        logger.debug(f'Starting search process with command "{" ".join(command)}".')
        try:
            proc = subprocess.Popen(
                command, encoding="ascii",
                stdout=search_log_f,  # if LOG_SEARCH is False, this is DEVNULL
                stderr=search_log_f,  # same as above
                preexec_fn=os.setsid                                            # TODO: setsid only POSIX; os comments suggest NOT TO USE THIS as it is unsafe against deadlocks. Suggestion is to use start_new_session keyword [UNIX only!] instead of setsid. For a Windows solution see https://stackoverflow.com/questions/47016723. Probably even better: do not use python subprocess, but QProcess (which can run with its own event loop, and has a .kill(), or perhaps .terminate()). QProcess may have some issues with OpenMPI v<=1.7 due to a bug there (see www.qtcentre.org/threads/19636-Qprocess-and-mpi-not-finishing).
                )
        except OSError:  # This should not fail unless the shell is very broken.
            logger.error("Error starting search. Check SD.TL file.")
            if search_log_path:
                search_log_f.close()
            raise
        else:
            pgid = os.getpgid(proc.pid)                                         # TODO: getpgid only POSIX
        if proc is None:
            logger.error("Error starting search subprocess... Stopping.")
            if search_log_path:
                search_log_f.close()
            raise RuntimeError("Error running search."
                               f'Could not start process using "{command}".')
        # FEED INPUT
        try:
            proc.communicate(input=rf_info_content, timeout=0.2)
        except subprocess.TimeoutExpired:
            pass  # started successfully; monitoring below
        except (OSError, subprocess.SubprocessError):
            logger.error("Error feeding input to search process. "
                         "Check files SD.TL and rf.info.")

        # MONITOR SEARCH
        searchStartTime = timer()
        last_debug_print_time = searchStartTime
        filepos = 0
        timestep = 1  # time step to check files
        # !!! evaluation time could be higher - keep low only for debugging; TODO
        evaluationTime = rp.searchEvalTime  # how often should SD.TL be evaluated # TODO: would be nicer with a QTimer, or even a QFileSystemWatcher
        lastEval = 0  # last evaluation time (s), counting from searchStartTime
        comment = ""
        sdtlGenNum = 0
        gaussianWidthOri = rp.GAUSSIAN_WIDTH
        check_datafiles = False
        try:
            while proc.poll() is None:  # proc is running
                time.sleep(timestep)
                parameters.update(rp)                                           # TODO: Would be way nicer with a QFileSystemWatcher
                # check convergence criteria
                stop = False
                checkrepeat = True
                if rp.STOP is True:
                    stop = True
                    checkrepeat = False
                    logger.info("Search stopped by STOP command.")
                    if not os.path.isfile("SD.TL"):
                        # try saving by waiting for SD.TL to be created...
                        logger.warning("SD.TL file not found. Trying to "
                                       "wait, maximum 5 minutes...")
                        i = 0
                        while not os.path.isfile("SD.TL") and i < 300:
                            time.sleep(1)
                            i += 1
                if rp.GAUSSIAN_WIDTH != gaussianWidthOri:
                    stop = True
                    repeat = True
                    comment = f"GAUSSIAN_WIDTH = {rp.GAUSSIAN_WIDTH}"
                    logger.info("GAUSSIAN_WIDTH parameter changed. "
                                "Search will restart.")
                t = timer() - searchStartTime
                if (t - lastEval > evaluationTime) or stop:
                    # evaluate
                    lastEval = t
                    newData = []
                    if os.path.isfile("SD.TL"):
                        filepos, content = iosearch.readSDTL_next(
                            offset=filepos
                            )
                        if content:
                            newData = iosearch.readSDTL_blocks(
                                content, whichR=rp.SEARCH_BEAMS,
                                n_expect=rp.SEARCH_POPULATION
                                )
                    elif t >= 900 and rp.HALTING < 3:                           # TODO: nicer with a QTimer timeout
                        stop = True
                        if tried_repeat:
                            logger.warning(
                                "No SD.TL file was written for 15 minutes "
                                "after restarting the search. Search will "
                                "stop."
                                )
                            repeat = False
                            rp.setHaltingLevel(2)
                        else:
                            repeat = True
                            tried_repeat = True
                            logger.warning(
                                "No SD.TL file was written for 15 minutes "
                                "after the search started. Trying to restart. "
                                "You can suppress this behaviour by setting "
                                "the HALTING parameter to 3."
                                )
                    for gen, rfacs, configs in newData:
                        gens.append(gen + genOffset)
                        sdtlGenNum = gen
                        rfaclist.append(np.array(rfacs))
                        dgen = {}
                        for k in ["dec", "best", "all"]:                        # TODO: would be nicer with a SearchConfigurationTracker class that has an .update
                            dgen[k] = gens[-1] - realLastConfigGen[k]
                        if configs != realLastConfig["all"]:
                            realLastConfig["all"] = configs
                            realLastConfigGen["all"] = gens[-1]
                        if (configs[:(rp.SEARCH_POPULATION // 10 + 1)]
                                != realLastConfig["dec"]):
                            realLastConfig["dec"] = configs[:(
                                            rp.SEARCH_POPULATION // 10 + 1)]
                            realLastConfigGen["dec"] = gens[-1]
                        if configs[0] != realLastConfig["best"]:
                            realLastConfig["best"] = configs[0]
                            realLastConfigGen["best"] = gens[-1]
                        for k in ["dec", "best", "all"]:
                            if (rp.SEARCH_MAX_DGEN[k] > 0
                                    and len(gens) > 1
                                    and not stop
                                    and rp.GAUSSIAN_WIDTH_SCALING != 1
                                    and dgen[k] >= rp.SEARCH_MAX_DGEN[k]):
                                stop = True
                                o = {"all": "all structures",
                                     "best": "best structure",
                                     "dec": "best 10% of structures"}
                                logger.info(
                                    "Search convergence criterion reached: "
                                    f"max. generations without change ({o[k]})"
                                    f": {dgen[k]}/{rp.SEARCH_MAX_DGEN[k]}."
                                    )
                                break
                    # decide to write debug info
                    # will only write once per SD.TL read
                    time_since_print = timer() - last_debug_print_time
                    current_gen = gens[-1] if gens else 0
                    if (current_gen - last_debug_write_gen > rp.output_interval):
                        speed = 1000*(timer() - absstarttime)/current_gen # in s/kG
                        logger.debug(
                            f"R = {min(rfacs)} (Generation {current_gen}, "
                            f"{time_since_print:.3f} s since "
                            f"gen. {last_debug_write_gen}, "
                            f"{speed:.1f} s/kG overall)"
                            )
                        last_debug_print_time = timer()
                        last_debug_write_gen = current_gen
                        check_datafiles = True
                    if newData:
                        _, _, lastconfig = newData[-1]
                    if (check_datafiles and not (stop and repeat)
                            and rp.PARABOLA_FIT["type"] != "none"):
                        check_datafiles = False
                        datafiles = [f for f in os.listdir()
                                     if re.match(r'data\d+\.chem$', f.lower())]
                        if rfac_predict and (rfac_predict[-1][1] >
                                             rfaclist[-1][0]):
                            # if current prediction is bad, reset x0
                            parab_x0 = None
                        try:
                            parab_x0, predictR = parabolaFit(
                                rp, datafiles, min(rfacs), x0=parab_x0,         # TODO: is rfacs correct here? may be undefined if the loop before does not run
                                max_configs=max_read_configs
                                )
                            if predictR is not None:
                                if not rfac_predict:
                                    logger.debug("Starting parabola fits "
                                                 "to R-factor data.")
                                rfac_predict.append((gens[-1], predictR))
                        except Exception:
                            logger.warning("Parabolic fit of R-factor "
                                           "data failed",
                                           exc_info=rp.is_debug_mode)
                    if len(gens) > 1:
                        try:
                            searchpdf.writeSearchProgressPdf(
                                rp, gens, rfaclist, lastconfig,
                                markers=markers, rfac_predict=rfac_predict
                                )
                        except Exception:
                            logger.warning("Error writing Search-progress.pdf",
                                           exc_info=rp.is_debug_mode)
                        try:
                            searchpdf.writeSearchReportPdf(rp)
                        except Exception:
                            logger.warning("Error writing Search-report.pdf",
                                           exc_info=rp.is_debug_mode)
                    if (len(gens) > 1 and os.path.isfile("SD.TL")
                            and (repeat or checkrepeat or not stop)):
                        try:
                            processSearchResults(sl, rp, search_log_path,
                                                 final=False)
                        except Exception as exc:                                # TODO: too general
                            logger.warning("Failed to update POSCAR_OUT "
                                           f"and VIBROCC_OUT: {exc}")
                if stop:
                    logger.info("Stopping search...")
                    kill_process(proc, default_pgid=pgid)
                    if (not repeat and not rp.GAUSSIAN_WIDTH_SCALING == 1
                            and checkrepeat):
                        repeat = True
                        block = False
                        for k in [k for k in ["dec", "best", "all"]
                                  if rp.SEARCH_MAX_DGEN[k] > 0]:
                            if convergedConfig[k] != realLastConfig[k]:
                                convergedConfig[k] = realLastConfig[k][:]
                                block = True
                            else:
                                repeat = False
                                o = {"all": "any structure",
                                     "best": "best structure",
                                     "dec": "best 10% of structures"}
                                logger.info(
                                    "Convergence reached: No improvement to "
                                    + o[k] + " since changing GAUSSIAN_WIDTH.")
                        if not repeat and block:
                            repeat = True
                            logger.info("Multiple convergence "
                                        "criteria are defined, not all are "
                                        "met. Search continues.")
                        if repeat:
                            if rp.GAUSSIAN_WIDTH <= 0.0001:
                                logger.info(
                                    "GAUSSIAN_WIDTH cannot be reduced "
                                    "further, continuing search...")
                            else:
                                rp.GAUSSIAN_WIDTH = max(
                                    rp.GAUSSIAN_WIDTH * rp.GAUSSIAN_WIDTH_SCALING,
                                    0.0001
                                )
                                logger.info(
                                    "Reducing GAUSSIAN_WIDTH parameter to {} "
                                    "and restarting search..."
                                    .format(round(rp.GAUSSIAN_WIDTH, 4)))
                                comment = ("GAUSSIAN_WIDTH = {}".format(
                                    round(rp.GAUSSIAN_WIDTH, 4)))
                            for k in ["dec", "best", "all"]:
                                rp.SEARCH_MAX_DGEN[k] = int(
                                            rp.SEARCH_MAX_DGEN[k] *
                                            rp.SEARCH_MAX_DGEN_SCALING[k])
                                realLastConfigGen[k] = gens[-1] if gens else 0
        except KeyboardInterrupt:                                               # TODO: would probably be nicer to install a custom signal handler for SIGINT, CTRL_C_EVENT, and CTRL_BREAK_EVENT, then place back the previous handler when done processing.
            if not os.path.isfile("SD.TL"):
                # try saving by waiting for SD.TL to be created...
                logger.warning("SD.TL file not found. Trying to wait, "
                               "interrupt again to override...")
                try:
                    i = 0
                    while not os.path.isfile("SD.TL") and i < 60:
                        time.sleep(1)
                        i += 1
                except KeyboardInterrupt:
                    pass   # user insisted, give up
            interrupted = True
            rp.STOP = True
            kill_process(proc, default_pgid=pgid)
            logger.warning("Search interrupted by user. Attempting "
                           "analysis of results...")
        except Exception:
            logger.error("Error during search. Check SD.TL file.")
            kill_process(proc, default_pgid=pgid)
            raise
        finally:
            # close open input and log files
            if search_log_path:
                search_log_f.close()
        if repeat:
            rp.SEARCH_START = "control"
            if gens:
                genOffset = gens[-1]
                rp.SEARCH_MAX_GEN -= sdtlGenNum
                markers.append((genOffset, comment))
            try:
                iosearch.generateSearchInput(sl, rp, steuOnly=True,
                                             cull=True, info=False)
            except Exception:
                logger.error("Error re-generating search input")
                raise
            if os.path.isfile("SD.TL"):
                try:
                    os.remove("SD.TL")
                except Exception:
                    logger.warning("Failed to delete old SD.TL file. "
                                   "This may cause errors in the "
                                   "interpretation of search progress.")
    if proc is not None:
        try:     # should generally not be necessary, but just to make sure
            kill_process(proc, default_pgid=pgid)
        except Exception:
            pass
    if not interrupted:
        logger.info("Finished search. Processing files...")
    else:
        logger.info("Processing files...")
    # final parabola fit  # !!! double check later
    if rfac_predict and (rfac_predict[-1][1] > rfaclist[-1][0]):
        parab_x0 = None
    datafiles = [f for f in os.listdir()
                 if re.match(r'data\d+\.chem$', f.lower())]
    if len(rfaclist) > 0:
        parab_x0, predictR = parabolaFit(rp, datafiles, min(rfacs),
                                         x0=parab_x0,
                                         max_configs=max_read_configs)
        if predictR is not None:
            rfac_predict.append((gens[-1], predictR))
    # write pdf one more time
    if len(gens) > 1:
        try:
            searchpdf.writeSearchProgressPdf(rp, gens, rfaclist, lastconfig,
                                             markers=markers,
                                             rfac_predict=rfac_predict)
        except Exception:
            logger.warning("Error writing Search-progress.pdf",
                           exc_info=True)
        try:
            searchpdf.writeSearchReportPdf(rp)
        except Exception:
            logger.warning("Error writing Search-report.pdf",
                           exc_info=True)
    # process SD.TL to get POSCAR_OUT, VIBROCC_OUT
    try:
        processSearchResults(sl, rp, search_log_path)
    except FileNotFoundError:
        logger.error("Cannot interpret search results without SD.TL file.")
        rp.setHaltingLevel(2)
    except Exception:
        logger.error("Error processing search results: ", exc_info=True)
        raise
    # if deltas were copied from domain folders, clean them up
    if rp.domainParams:
        rgx = re.compile(r'D\d+_DEL_')
        for file in [f for f in os.listdir() if rgx.match(f)]:
            try:
                os.remove(file)
            except Exception:
                logger.warning('Failed to deleted redundant domain delta file '
                               + file)
    # process files
    try:
        os.rename('PARAM', 'search-PARAM')
    except Exception:
        logger.warning("Failed to rename search input file PARAM to "
                       "search-PARAM")
    try:
        os.rename('rf.info', 'search-rf.info')
    except Exception:
        logger.warning("Failed to rename search input file rf.info to "
                       "search-rf.info")
    if lastconfig is not None:
        rp.searchResultConfig = lastconfig



"""ViPErLEED utility: Delete atoms below a certain height."""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Florian Kraushofer (@fkraushofer)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-08-03'
__license__ = 'GPLv3+'

from viperleed.cli_base import float_in_zero_one
from viperleed.utilities.poscar.base import _RemoveAtomsCLI


class DeleteBelowCLI(_RemoveAtomsCLI, cli_name='delete_below'):
    """Remove atoms below a certain c fraction."""

    long_name = 'delete atoms below'

    def add_remove_condition_arguments(self, parser):
        """Add maximum c fraction to parser."""
        parser.add_argument(
            'c',
            help='delete all atoms strictly below this c fraction',
            type=float_in_zero_one,
            )

    def process_slab(self, slab, args):
        """Remove atoms below the c fraction specified in the CLI."""
        modified_slab = super().process_slab(slab, args)
        logger = self.get_logger()
        logger.debug(f'Deleted {slab.n_atoms - modified_slab.n_atoms} '
                     f'atoms at c < {args.c}.')
        return modified_slab

    def select_surviving_atoms(self, slab, args):
        """Return atoms of slab with c position of at least args.c."""
        return [atom for atom in slab if atom.pos[2] >= args.c]


if __name__ == '__main__':
    DeleteBelowCLI.run_as_script()



"""
=======================================
   ViPErLEED Graphical User Interface
=======================================
 *** module guilib.leedsim.widgets ***

Created: 2020-01-12
Author: Michele Riva

Blah blah TODO
"""

# TESTING
from time import perf_counter
# TESTING

import numpy as np
import matplotlib as mpl
#OR
#from matplotlib.patches import Circle as MplCircle
from matplotlib import colors as mpl_colors
import PyQt5.QtCore as qtc
import PyQt5.QtGui as qtg
import PyQt5.QtWidgets as qtw

# import guilib as gl
from viperleed import guilib as gl

TEST = False

class RealCanvas(gl.MPLFigureCanvas):
    
    def plotLattices(self):
        self.ax.cla()
        
        _win = self.window()  # this should be the LEED_GUI instance
        # if not isinstance(_win, gl.LEED_GUI):  # REMOVED 2020-01-25
            # raise RuntimeError("I expected a LEED_GUI instance, "
                               # f"but I found {_win}")
        
        rs = _win.real
        
        rotation = _win.get_angle()
        
        #### BULK #### 
        rotBulkB = rs.bulk.get_rotated_basis(rotation)
        
        # Plot bulk as lines. A line is p+x*v, where p is any point along 
        # the line, and v a vector parallel to the line
        alpha = 0.15
        lineColor = tuple(alpha*fg + (1-alpha)*bg
                          for (fg, bg) in zip(mpl_colors.to_rgb('gray'),
                                              mpl_colors.to_rgb('white')))
        
        x = np.array([-rs.fov, rs.fov])
        K = max(rs.bulk.hk.ravel())
        # the following is a fast code to plot lines, adapted from 
        # http://exnumerus.blogspot.com/2011/02/
        #                              /how-to-quickly-plot-multiple-line.html
        xlist = []
        ylist = []
        for i in [0, 1]:
            slope = [x*v for v in rotBulkB[i]]
            for k in range(-K, K):
                xlist.extend(slope[0] + k*rotBulkB[(i+1)%2, 0])
                xlist.append(None)
                ylist.extend(slope[1] + k*rotBulkB[(i+1)%2, 1])
                ylist.append(None)
        self.ax.plot(xlist, ylist, color=lineColor, zorder=-1)
        for unitVec in rotBulkB:
            self.ax.annotate("", xy=tuple(unitVec),
                        xytext=(0, 0),
                        arrowprops=dict(arrowstyle="->", color='gray')
                        )
        
        #### SURFACE ####
        windowScaleFactor = (self.rightSize().height()/434)**2
                            # this is used for scaling the size of the spots
        rotSurfB = rs.surf.get_rotated_basis(rotation)
        rotSurfL = rs.surf.get_rotated_lattice(rotation)
        
        # Plot surface as scatter
        self.ax.scatter(rotSurfL[:,0], rotSurfL[:,1],
                        s=8*windowScaleFactor, c='k', zorder=2)
        for unitVec in rotSurfB:
            self.ax.annotate("", xy=tuple(unitVec),
                             xytext=(0, 0),
                             arrowprops=dict(arrowstyle="->"))
        
        self.setAxLimits(rs.fov)
        self.ax.figure.canvas.draw_idle()


class LEEDCanvas(gl.MPLFigureCanvas):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
        # Get rid of the spines completely in the LEED pattern
        self.setSpinesOn(False)
        
        # Define annotations that will appear when the mouse hovers above
        # LEED spots (including symmetry-equivalent). Notice that at most 12
        # symmetry-equivalent spots can exist at the same time
        self.annots = [gl.HoverAnnot(self.parentWidget()) for _ in range(12)]
        
        if TEST:
            self.mpl_connect('motion_notify_event', self.on_hover)
        
        # ## TESTING
        # self.mpl_connect('figure_enter_event', self.enter_figure)
        # self.mpl_connect('figure_leave_event', self.leave_figure)
        # self.mpl_connect('button_press_event', self.on_mousepress)
        
        tmpoffs = 0
        deltas = [np.array([np.cos(x + tmpoffs), np.sin(x + tmpoffs)])
                  for x in np.linspace(0, 2*np.pi,
                                       num=len(self.annots), endpoint=False)]
        for (annot, delta) in zip(self.annots, deltas):
            annot.head_pos = delta*10 + annot.center
        # ## END TESTING
        
    def initLEEDScreen(self, lineThick=0.1):
        # initialize LEED screen to a radius of 1
        if not isinstance(lineThick, (int, float)):
            raise ValueError("initLEEDScreen requires a non-negative"
                             " floating point input")
        elif lineThick < 0:
            raise ValueError("line thickness cannot be negative")
        else:
            screen = mpl.patches.Circle((0, 0), 1,  facecolor='white',
                                        edgecolor='k', lw=lineThick,
                                        zorder=-1)

        self.ax.add_patch(screen)
        self.setAxLimits(1)
        self.screen = screen
        
    def resetLEEDCanvas(self):
        self.ax.cla()  # clear the axes
        self.setSpinesOn(False)  # turn off again the spines
        self.ax.add_patch(self.screen)  # and add once more the screen circle
        
        _win = self.window()

        energy = _win.get_energy()  # read energy from input
        self.rotation = _win.get_angle()
        
        self.leed = _win.leed
        
        screenR = self.leed.screen_radius(energy)
        self.screen.set_radius(screenR)
        self.setAxLimits(screenR)
        
        windowScaleFactor = (self.rightSize().height()/434)**2
        
        # scaling factor for the markers, also depends on the window size
        self.markScale = np.sqrt(self.leed.max_energy/energy)*windowScaleFactor
        
        # Big dot at (0,0), on top of everything
        self.ax.scatter(0, 0, s=100*self.markScale, c='k', zorder=1e10)
        
        # plot the bulk spots as hollow circles
        bulkLEED = self.leed.rotate(self.rotation, 'bulk')
        self.bulkLEED = self.ax.scatter(bulkLEED[:,0], bulkLEED[:,1],
                                        s=25*self.markScale,
                                        facecolors='none',
                                        edgecolors='k')
        
        # Now clip the pattern with the circle
        self.bulkLEED.set_clip_on(True)
        self.bulkLEED.set_clip_path(self.screen)
    
    def plotLEED(self, hideDoms):
        self.resetLEEDCanvas()
        self.leed.rotate(self.rotation, 'surf')
        
        if hideDoms or self.leed.n_domains == 1:
            toPlot = self.leed.firstLEED
        else:
            toPlot = self.leed.domsLEED
        
        self.leedPat = []  # list of scatter plots
        for patt in toPlot:
            self.leedPat.append(
                            self.ax.scatter(patt.rotBeams[:,0],
                                            patt.rotBeams[:,1],
                                            s=5*self.markScale*patt.sizeScale,
                                            c=patt.color, marker=patt.marker
                                            )
                               )
        for patt in self.leedPat:
            patt.set_clip_on(True)
            patt.set_clip_path(self.screen)
            # Set pick radius for hovering annotations (scale with marker size).
            # Whenever the mouse is closer to a spot than pickradius points, the
            # mouse event will be considered 'contained' in the pattern to which
            # the spot belongs
            patt.set_pickradius(3*np.sqrt(self.markScale))
        self.ax.figure.canvas.draw_idle()
    
    def showEvent(self, event):
        super().showEvent(event)
        
        # Now set the center of the LEED screen as the 'center' of annotations
        # (i.e., annotations will always be radially symmetric around the screen
        # center)
        # NB: this needs to be UPDATED when I can handle off-normal incidence,
        #     as the center should then be the (0, 0) spot!
        #     It's most likely better to move/replicate this part in the
        #     resetLEEDCanvas as it will need to know where the (0, 0) spot is
        local_center = self.geometry().center()
        global_center = self.parentWidget().mapToGlobal(local_center)
        for annot in self.annots:
            annot.center = global_center
    
    def on_hover(self, event):
        if not self.is_on_screen(event):
            [annot.hide() for annot in self.annots]
            return
        show_annots = False
        for patt in self.leedPat:
            contained, ind = patt.contains(event)
            if contained:
                show_annots = True
                mpl_pos = self.closest_point(
                    self.scatter_coords(patt, ind['ind']),
                    (event.x, event.y))
                qt_pos = self.mpl_to_qt_coordinates(*mpl_pos)
                
                # TESTING
                self.annots[0].head_pos = self.mapToGlobal(qt_pos)
                # end TESTING
        
        for annot in self.annots:  # needs to select only those that have to be
                                   # shown
            annot.update_position()
            annot.setVisible(show_annots)
            annot.update()
    
    def is_on_screen(self, event):
        """
        Checks if the mouse event occurred inside the LEED screen.
        """
        # in principle it looks like one could use the matplotlib patch
        # .contains method to do this, but for reasons I don't understand it
        # seems to return True as soon as the mouse enters the bounding box of
        # the screen rather than its interior
        if event.inaxes != self.ax:
            return False
        event_r = np.sqrt(event.xdata**2 + event.ydata**2)
        return event_r <= self.screen.radius
    
    def scatter_coords(self, scatt_plot, indices):
        """
        Given a scatter plot and a list of indices it returns the (x, y)
        coordinates of the scatter points at the corresponding indices. The
        coordinates are given in the reference system of the canvas.
        """
        to_restore = scatt_plot.get_offset_position()
        scatt_plot.set_offset_position('data')
        mpl_pos = self.ax.transData.transform(
                    scatt_plot.get_offsets()[indices])
        scatt_plot.set_offset_position(to_restore)
        
        return mpl_pos
    
    def closest_point(self, points, close_to):  # move to base
        """
        Out of the array-like 'points' returns the one that is closest to
        'close_to'
        """
        dist = np.linalg.norm(np.subtract(points, close_to), axis=1)
        closest = np.where(abs(dist - dist.min()) <= 1e-8)
        
        return np.asarray(points)[closest][0]

    # TESTING
    if TEST:
        def enter_figure(self, event):
            for annot in self.annots:
                annot.show()
        
        def leave_figure(self, event):
            for annot in self.annots:
                annot.hide()
        
        def on_mousepress(self, event):
            x_qt, y_qt = self.mpl_to_qt_coordinates(event.x, event.y)
            print(f'(x, y) = MPL: ({event.x}, {event.y}), Qt: ({x_qt}, {y_qt})')
    # end TESTING
    



class RotationBlock(gl.TextBoxWithButtons):
    def __init__(self, parent=None):
        params = {
            'labelText': '&Rotation (\u00b0)',
            'textBoxText': '\u2014',
            'botButText': '\u21bb', #CW
            'topButText': '\u21ba', #CCW
            'parent': parent,
            'textBoxTip': 'Change rotation of lattices and pattern. '\
                          'Positive angles are counterclockwise',
            'topButTip': 'Rotate lattices and pattern 10° clockwise'\
                         'Hold Ctrl down for finer control',
            'botButTip': 'Rotate lattices and pattern 10° counterclockwise'\
                         'Hold Ctrl down for finer control',
            'textBoxWidth': 85
        }
        
        super().__init__(**params)
        self.setTips(**params)
        self.text.setLimits(-180, 180, 'cyclic')
        self.text.setStep(10, 'add')
        
        #set aliases for easier identification
        self.ccw = self.topBut
        self.cw = self.botBut
        
        #and build the bottom part
        self.makeBottomWidget()
    
    def makeBottomWidget(self):
        # bottomWidget looks like this:
        #
        #    'Align bulk'
        #     [1 0] hor ver
        #     [0 1] hor ver
        #
        # will be composed with a QGridLayout, that is then assigned to self.bottomWidget
        
        bWLay = qtw.QGridLayout()
        bWLay.setSpacing(0)
        bWLay.setContentsMargins(0, 0, 0, 0)
        
        # prepare the sub-widgets
        rotBulkLab = qtw.QLabel('Align bulk:')
        oneZeroLab = qtw.QLabel("[1 0]")
        zeroOneLab = qtw.QLabel("[0 1]")
        self.h10 = qtw.QPushButton("Hor.")
        self.v10 = qtw.QPushButton("Ver.")
        self.h01 = qtw.QPushButton("Hor.")
        self.v01 = qtw.QPushButton("Ver.")
        horVerButs = ((self.h10, self.v10), (self.h01, self.v01))
        
        # add the relevant ones to the list of sub-widgets
        self.subWidgs.extend([self.h10, self.v10, self.h01, self.v01,
                             *np.ravel(horVerButs)])
        
        # set fonts
        [lab.setFont(gl.AllGUIFonts().smallTextFont)
         for lab in [rotBulkLab,oneZeroLab,zeroOneLab]]
        [but.setFont(gl.AllGUIFonts().smallButtonFont)
         for but in np.ravel(horVerButs)]
        
        # set size policies and adjust the sizes to account for the new font
        for lab in [rotBulkLab, oneZeroLab, zeroOneLab]: 
            lab.setSizePolicy(qtw.QSizePolicy.Fixed, qtw.QSizePolicy.Preferred)
            lab.adjustSize()
        for but in np.ravel(horVerButs):
            but.setSizePolicy(qtw.QSizePolicy.Preferred, qtw.QSizePolicy.Fixed)
            but.setMaximumHeight(qtw.QPushButton().sizeHint().height()*.84)
        
        # set tooltips
        tips = ['Rotate lattices and pattern to bring the [1 0] '
                    + 'bulk vector horizontal',
                'Rotate lattices and pattern to bring the [1 0] '
                    + 'bulk vector vertical',
                'Rotate lattices and pattern to bring the [0 1] '
                    + 'bulk vector horizontal',
                'Rotate lattices and pattern to bring the [0 1] '
                    + 'bulk vector vertical']
        for (but, tip) in zip(np.ravel(horVerButs), tips):
            but.setStatusTip(tip)
            but.setToolTip(tip)
        
        # now add the widgets to the layout
        bWLay.addWidget(rotBulkLab, 0, 0, 1, 3, qtc.Qt.AlignLeft)
        bWLay.addWidget(oneZeroLab, 1, 0, 1, 1,
                        qtc.Qt.AlignHCenter | qtc.Qt.AlignLeft)
        bWLay.addWidget(zeroOneLab, 2, 0, 1, 1,
                        qtc.Qt.AlignHCenter | qtc.Qt.AlignLeft)
        [bWLay.addWidget(horVerButs[i][j], i+1, j+1, 1, 1, qtc.Qt.AlignCenter)
         for i in range(2) for j in range(2)]
        
        # set the minimum heights of rows and columns
        bWLay.setColumnMinimumWidth(0, oneZeroLab.width()*1.2)
        
        # Add the layout to the bottomWidget
        self.bottomWidget.setLayout(bWLay)
        self.bottomWidget.layout().activate()
    
    def on_botWidgPressed(self, pressed):
        direction = 0
        vert = False
        if pressed in (self.h01, self.v01):
            direction = 1
        if pressed in (self.v10, self.v01):
            vert = True
        angle = self.window().real.angle_for_horizontal_bulk(direction)
        if vert:
            angle += 90
        if qtw.qApp.keyboardModifiers() == qtc.Qt.ControlModifier:
            angle += 180
        self.text.updateText(angle)


class EnergyBlock(gl.TextBoxWithButtons):
    def __init__(self, parent=None):
        params = {'labelText': '&Energy (eV)',
                  'textBoxText': '\u2014',
                  'topButText': '\u25b2',  # energy up
                  'botButText': '\u25bc',  # energy down
                  'parent': parent,
                  'textBoxTip': 'Change primary energy of the electron beam',
                  'topButTip': 'Increase energy by 20%. ' \
                               'Hold Ctrl down for a 5% increase',
                  'botButTip': 'Decrease energy by 20%. ' \
                               'Hold Ctrl down for a 5% decrease',
                  'textBoxWidth': 85
                  }

        super().__init__(**params)
        self.setTips(**params)
        self.text.setStep(1.2, 'scale')
        
        #set aliases for easier identification
        self.enUp = self.topBut
        self.enDown = self.botBut
        
        # for reasons unknown the down arrow is typeset weirdly
        # (larger than it should) -> use smaller size
        self.enDown.setFont(gl.AllGUIFonts().smallButtonFont)
        
        self.makeBottomWidget()
    
    def makeBottomWidget(self):
        # The bottomWidget in this case contains one QlineEdit with
        # two text lines that state the minimum and maximum energy
        #
        self.limits = qtw.QLabel('Min = 10 eV\nMax = \u2014')
        self.limits.setFont(gl.AllGUIFonts().smallTextFont)
        self.limits.adjustSize()
        
        bwLay = qtw.QHBoxLayout()
        bwLay.setSpacing(0)
        bwLay.setContentsMargins(0,0,0,0)
        bwLay.addWidget(self.limits, qtc.Qt.AlignLeft | qtc.Qt.AlignTop)
        
        self.bottomWidget.setLayout(bwLay)
        self.subWidgs.extend([self.limits])
        
        self.layout().setRowMinimumHeight(5, self.limits.height())


class DomsBlock(qtw.QWidget):
    toggleText = ['Hide &Domains', 'Show &Domains']
    # These are the two possible strings that appear on the toggle button.
    # Pressing alt+d fires the toggle button
    
    def __init__(self, parent=None):
        super().__init__(parent)
        self.hide = False
        
        # The next line is useful only if I ever want to place this in a 
        # layout
        self.setSizePolicy(qtw.QSizePolicy.Minimum, qtw.QSizePolicy.Minimum)
        self.text = qtw.QLabel('\u2014 inequivalent domain(s)')
        self.toggle = ToggleButton(self.toggleText[self.hide])
        
        self.subWidgs = (self.text, self.toggle)
        
        self.setTips(text='Open or drag-drop a LEED pattern file to'
                          'determine how many domains!',
                     but='Toggle visibility of symmetry-equivalent domains')
        self.compose()
    
    def setTips(self, **kwargs):
        if 'text' in kwargs.keys():
            self.text.setStatusTip(kwargs['text'])
        if 'but' in kwargs.keys():
            tip = kwargs['but']
            self.toggle.setToolTip(tip)
            self.toggle.setStatusTip(tip)
    
    def compose(self):
        #set fonts
        self.text.setFont(gl.AllGUIFonts().labelFont)
        self.toggle.setFont(gl.AllGUIFonts().buttonFont)
        
        #set sizes
        self.toggle.setSizePolicy(qtw.QSizePolicy.Minimum,
                                  qtw.QSizePolicy.Minimum)
        self.toggle.adjustSize()
        
        self.text.setAlignment(qtc.Qt.AlignRight | qtc.Qt.AlignBaseline)
        self.text.setSizePolicy(qtw.QSizePolicy.Expanding,
                                qtw.QSizePolicy.Minimum)
        self.text.adjustSize()
        self.text.setMinimumSize(self.text.size())
        self.toggle.setMinimumSize(self.toggle.size())
        
        #----- Set up the layout -----#
        lay = qtw.QVBoxLayout()
        lay.setSpacing(2)
        lay.setContentsMargins(0, 0, 0, 0)
        
        lay.addStretch(1)  # the stretches keep things together
        lay.addWidget(self.text) 
        lay.addWidget(self.toggle)
        lay.addStretch(1)
        lay.setAlignment(self.text, self.text.alignment())
        lay.setAlignment(self.toggle, qtc.Qt.AlignRight | qtc.Qt.AlignVCenter)
        
        self.setLayout(lay)
        
        self.adjustSize()
    
    def width(self):
        self.ensurePolished()
        return max(sub.width() for sub in self.subWidgs)
    
    def height(self):
        self.ensurePolished()
        return (sum(sub.height() for sub in self.subWidgs) 
                + self.layout().spacing())
    
    def minimumSizeHint(self):
        return self.sizeHint()
    
    def sizeHint(self):
        return qtc.QSize(self.width(), self.height())
    
    def togglePressed(self):
        self.hide = not self.hide
        self.toggle.setText(self.toggleText[self.hide])
        self.matricesPopup.updateMatrices(self.hide)
    
    def updateText(self, txt):
        if not isinstance(txt, str):
            raise
        
        self.text.setText(txt)
        self.text.setMinimumSize(self.text.sizeHint())
        
        trOld = self.geometry().topRight()
        self.adjustSize()
        newPos = (self.geometry().topLeft()
                  - self.geometry().topRight()
                  + trOld)
        self.move(newPos)

    def initPopup(self):
        leed = self.window().leed
        self.matricesPopup = MatricesPopup(leed.superlattices,
                                           leed.domColors,
                                           parent=self.window())
    
    #----------------------- Re-implement mouseEvents -----------------------#
    def mouseReleaseEvent(self, event):
        if self.underMouse():
            child = self.childAt(event.pos())
            if child == self.text:
                # Position and show matricesPopup
                if not self.matricesPopup.shown:
                    # not shown before --> place it at the standard position
                    tR = self.mapToGlobal(self.text.geometry().topRight())
                    sizePopup = self.matricesPopup.frameSize()
                    newPos = qtc.QPoint(tR.x() - 0.7*sizePopup.width(),
                                    tR.y() - sizePopup.height() - 25)
                    self.matricesPopup.move(newPos)
                    self.matricesPopup.dragPosition = newPos
                    self.matricesPopup.shown = True
                if self.matricesPopup.isVisible():
                    self.matricesPopup.hide()
                else:
                    self.matricesPopup.show()
        super().mouseReleaseEvent(event)


class ToggleButton(qtw.QPushButton):
    def sizeHint(self):
        self.ensurePolished()
        refbutton = qtw.QPushButton(self.text())
        refbutton.setFont(self.font())
        return refbutton.sizeHint()*1.2
    
    def minimumSizeHint(self):
        return self.sizeHint()


class MatricesPopup(qtw.QWidget):
    def __init__(self, matrices, colors,
                 fs=gl.AllGUIFonts().mathFont.pointSize(),
                 parent=None):
        self.parent = parent
        super().__init__()
        
        if colors is not None:  # there's more than one domain
            self.matrices = [gl.PainterMatrix(matrix, color=color, fs=fs)
                             for (matrix, color) in zip(matrices, colors)]
            # add one matrix to be displayed in black when the user toggles 
            # the visibility of the domains
            self.firstDomMatrix = gl.PainterMatrix(matrices[0], fs=fs)
        elif len(matrices) == 1:
            self.matrices = [gl.PainterMatrix(matrices[0], fs=fs)]
        else:  # some input error: colors are missing
            raise
        
        self.shown = False  # used for choosing whether is should pop up at 
                            # the standard position or not
        self.initPopup()
    
    def initPopup(self):
        # The next line makes the widget not steal focus from the main 
        # window when it is shown
        self.setAttribute(qtc.Qt.WA_ShowWithoutActivating)
        self.setParent(self.parent)  # This should come before the next 
                                     # lines to have the window stay on top 
                                     # of the main, but not on top of other 
                                     # applications
        
        #-------- Edit the appearance of the window --------#
        # Qt.Tool: thinner frame
        # Qt.CustomizeWindowHint: Remove titlebar, icon, and 
        #                         maximize/minimize/close buttons
        # Qt.WindowStaysOnTopHint: always keep it on top - do not use. 
        #                          Rather parent it to the MainWindow so 
        #                          that it stays only on top of it (and not 
        #                          on top of other apps)
        # Qt.WindowDoesNotAcceptFocus: prevent switching focus to the popup
        self.setWindowFlags(qtc.Qt.Tool
                            | qtc.Qt.CustomizeWindowHint
                            | qtc.Qt.WindowDoesNotAcceptFocus)
        gl.editStyleSheet(self, 'background-color: white;')
        
        # The next attribute is used in the mousePressEvent and 
        # mouseMoveEvent for dragging the window around
        self.dragPosition = self.frameGeometry().topLeft()
        
        #--------- Prepare layout for the matrices ---------#
        nMatrices = len(self.matrices)  # this is 1, 2, 3, 4, 6, 8, or 12
                                        # TRUE?
        if nMatrices in range(1, 4):  # 1 -- 3
            nRows = 1
        elif nMatrices in range(4, 10):  # 4 -- 8
            nRows = 2
        else:
            nRows = 3
        nCols = int(nMatrices/nRows)
        
        lay = qtw.QGridLayout()
        [lay.addWidget(matrix, index//nCols, index % nCols, qtc.Qt.AlignCenter)
         for (index, matrix) in enumerate(self.matrices)]
        
        if hasattr(self, 'firstDomMatrix'):
            # More than one domain. Add anyway the matrix of the first domain
            # for changing its color when domains are toggled. Initially 
            # hide (toggling the button will show it).
            lay.addWidget(self.firstDomMatrix, 0, 0, qtc.Qt.AlignCenter) 
            self.firstDomMatrix.hide()
        
        sizes = [(matrix.sizeHint().width(), matrix.sizeHint().height())
                 for matrix in self.matrices]
        (widths, heights) = zip(*sizes)
        
        [lay.setColumnMinimumWidth(col, max(widths))
         for col in range(lay.columnCount())]
        [lay.setRowMinimumHeight(row, max(heights))
         for row in range(lay.rowCount())]
        lay.setSizeConstraint(qtw.QLayout.SetFixedSize)  # This makes the window 
                                                         # not re-sizable
        self.setLayout(lay)
        self.layout().activate()
    
    def mousePressEvent(self, event):
        if event.button() == qtc.Qt.LeftButton:
            self.dragPosition = (event.globalPos()
                                - self.frameGeometry().topLeft())
    
    def mouseReleaseEvent(self, event):
        event.accept()
    
    def mouseDoubleClickEvent(self, event):
        event.accept()
    
    def mouseMoveEvent(self, event):
        if event.buttons() == qtc.Qt.LeftButton:
            self.move(event.globalPos() - self.dragPosition)
    
    def updateMatrices(self, domsHidden):
        if domsHidden:
            toHide = self.matrices[0]
            toShow = self.firstDomMatrix
        else:
            toHide = self.firstDomMatrix
            toShow = self.matrices[0]
        toHide.hide()
        toShow.show()


class HoverAnnot(qtw.QWidget):
    # Arrow properties:
    headW = 6  # total width of arrowhead
    headL = 1.2*headW  # total length of arrowhead
    arrowL = 40  # total length of arrow segment
    
    # quantities used for painting the frame around the text
    framePad = (5, 3)  # amount of space added around the text
    
    # Painting styles and colors:
    bgColor = qtg.QColor(qtc.Qt.white)
    bgColor.setAlphaF(0.8)
    lineColor = qtg.QColor.fromHsvF(0, 0, 0.15)  # somewhat gray
    lineWidth = 1.5
    gap = 3 + lineWidth
    
    paintPen = qtg.QPen()
    paintPen.setWidthF(lineWidth)
    paintPen.setBrush(lineColor)
    paintPen.setCapStyle(qtc.Qt.RoundCap)
    paintPen.setJoinStyle(qtc.Qt.RoundJoin)
    
    def __init__(self, parent=None, text='Test some longer text', center=(500, 500)):
        super().__init__()
        
        self.setAttribute(qtc.Qt.WA_ShowWithoutActivating)
        self.setAttribute(qtc.Qt.WA_TranslucentBackground)
        # With the next attribute, even if the widget is too large mouse
        # events will be passed on to the widgets directly below it (spatially)
        self.setAttribute(qtc.Qt.WA_TransparentForMouseEvents)
        self.setParent(parent)
        
        self.setWindowFlags(qtc.Qt.Tool
                            | qtc.Qt.CustomizeWindowHint
                            | qtc.Qt.FramelessWindowHint
                            | qtc.Qt.WindowDoesNotAcceptFocus)
        
        self.setFont(gl.AllGUIFonts().labelFont)

        # Global position of the point that, together with the arrow head_pos
        # defines the direction of the arrow (from head_pos to center)
        self.__center = center
        
        self.__head_pos = None  # Global position of the arrowhead
        self.__tip_to_center = None  # distance from tip to center of text box
        
        self.__text = ''
        self.frame = qtg.QPainterPath()  # Frame that will contain the text,
        self.framePos = qtc.QPoint()     # its local position (top-left),
        self.frameRadius = 0             # the curvature radius of its corners
        self.textBbox = qtc.QRectF()     # and the boundingRect of text
        
        self.text = text
    
    @property
    def center(self):
        """
        Global position of the point towards which the arrow points
        """
        return self.__center
    
    @center.setter
    def center(self, ctr):
        """
        Set the global position of the point towards which the arrow points to
        ctr
        """
        if (not gl.check_type(ctr, 'arraylike')
            and not isinstance(ctr, (qtc.QPoint, qtc.QPointF))):
            raise TypeError("head_pos must be array-like, QPoint or QPointF."
                            f"Found {type(ctr)} instead.")
        if isinstance(ctr, (qtc.QPoint, qtc.QPointF)):
            ctr = (ctr.x(), ctr.y())
        
        try:  # this accounts for the case in which head_pos is not defined yet
            delta_head = self.head_pos - self.center
        except TypeError:
            delta_head = None
        self.__center = np.asarray(ctr)
        
        if delta_head is not None:
            self.head_pos = self.center + delta_head
    
    @property
    def text(self):
        return self.__text
    
    @text.setter
    def text(self, text):
        if not isinstance(text, str):
            raise
        self.__text = text
        self.updateFrameOutline()
        self.updateFramePosition()
    
    @property
    def head_pos(self):
        """
        Global position of the arrowhead
        """
        return self.__head_pos
    
    @head_pos.setter
    def head_pos(self, pos):
        """
        Set the global position of the arrowhead to pos
        """
        if (not gl.check_type(pos, 'arraylike')
            and not isinstance(pos, (qtc.QPoint, qtc.QPointF))):
            raise TypeError("head_pos must be array-like, QPoint or QPointF."
                            f"Found {type(pos)} instead.")
        if isinstance(pos, (qtc.QPoint, qtc.QPointF)):
            pos = (pos.x(), pos.y())
        
        if not len(pos) == 2:
            raise ValueError("Invalid position. An array-like with length == "
                             f"2 is needed. Found {len(pos)} instead.")

        self.__head_pos = np.asarray(pos)

        self.updateFramePosition()
    
    @property
    def tip_pos(self):
        """
        Returns the correct local position of the tip of the arrow such that
        the whole annotation can be painted within the area of the widget.
        Painting of the arrow will start at this position.
        """
        theta = np.radians(self.angle)
        sin_theta = np.sin(theta)
        cos_theta = np.cos(theta)
        frame_h = self.frame.boundingRect().height() + self.lineWidth
        frame_w = self.frame.boundingRect().width() + self.lineWidth

        tip_to_center = self.__tip_to_center + self.lineWidth
        
        delta_h = frame_h/2 + tip_to_center*sin_theta
        delta_w = frame_w/2 + tip_to_center*cos_theta
        
        # set the minimum delta to 2 to take care of the arrowhead triangle
        # +1 px for anti-alias, and one more pixel on h to avoid issues when the
        # arrow is in the bottom-left quadrant and exiting from the short side
        delta_w = max(2, delta_w) + 1
        delta_h = max(2, delta_h) + 2
        
        # notice that I return rounded values so that, later on, the widget can
        # be moved correctly to this position: only int values are acceptable
        # for widget movements
        return qtc.QPointF(np.round(delta_w), np.round(delta_h))
    
    @property
    def direction(self):
        try:
            dir_vec = self.center - self.head_pos
        except TypeError:
            # one of the two above is undefined
            return None
        return dir_vec/np.linalg.norm(dir_vec)
    
    @property
    def angle(self):
        """
        Returns the angle in degrees
        """
        dir_vec = self.direction
        if dir_vec is None:
            return None
        return np.degrees(np.arctan2(dir_vec[1], dir_vec[0]))
    
    def sizeHint(self):
        self.ensurePolished()
        rect_size = self.frame.boundingRect().size()
        delta = self.__tip_to_center + self.lineWidth
        w = np.round(rect_size.width())
        h = np.round(rect_size.height())
        # notice that the return value is slightly over-sized even in the worst
        # case scenario, as one would strictly need to add only w/2 and h/2
        return qtc.QSize(w + delta, h + delta)
    
    def minimumSizeHint(self):
        return self.sizeHint()
    
    def updateFrameOutline(self):
        self.frame = qtg.QPainterPath()
        self.textBbox = qtg.QFontMetricsF(self.font()).boundingRect(self.text)
        
        bbox = qtc.QRectF(self.textBbox)
        bbox.adjust(-self.framePad[0], -self.framePad[1],
                     self.framePad[0], self.framePad[1])
        bbox.moveTo(self.framePad[0]/2, self.framePad[1]/2)
        self.frameRadius = 0.2*bbox.height()
        self.frame.addRoundedRect(bbox, self.frameRadius, self.frameRadius)
        
        self.textBbox.adjust(-0.5, -0.5, 0.5, 0.5)
        self.textBbox.moveTo(self.framePad[0] + 1, self.framePad[1] + 1)
    
    def updateFramePosition(self):
        # Get the right position for self.frame such that the center of the 
        # text box is coincident with the prolongation of the arrow
        
        if self.angle is not None:
            bbox = self.frame.boundingRect()
            w = bbox.width()
            h = bbox.height()
            
            if np.abs(np.sin(np.radians(self.angle))) > np.radians(1):
                # away from the horizontal direction by more than 1 degree
                dy = -h/2 * np.sign(self.direction[1])
            else:
                dy = 0
            
            if dy == 0:
                dx = -w/2 * np.sign(self.direction[0])
            elif np.abs(np.cos(np.radians(self.angle))) > np.radians(1):
                # away from the vertical direction by more than 1 degree
                dx = dy / np.tan(np.radians(self.angle))
            else:
                dx = 0
            
            if np.abs(dx) > w/2:
                # Shallow horizontal angle, the arrow should come out from the
                # vertical edge rather than from the horizontal one
                dx = -w/2 * np.sign(self.direction[0])
                dy = dx * np.tan(np.radians(self.angle))
            
            # Set up the gap between the arrow and the outer edge of the box, 
            # including a correction for the rounded corner
            gap = self.gap  # takes into account line width
            tt = np.abs(np.tan(np.radians(self.angle)))
            ct = np.abs(np.cos(np.radians(self.angle)))
            st = np.abs(np.sin(np.radians(self.angle)))
            r = self.frameRadius
            if tt > max((h - 2*r)/w, h/(w - 2*r)):
                (ddx, ddy) = (0.5*gap*ct, 0.5*gap*(1 + st))
            elif tt < min((h - 2*r)/w, h/(w - 2*r)):
                (ddx, ddy) = (0.5*gap*(1 + ct), 0.5*gap*st)
            else:  # arrow intersects at rounded corner -> different gap
                (ddx, ddy) = (0.8*gap*ct, 0.8*gap*st)
            
            dx -= ddx*np.sign(self.direction[0])
            dy -= ddy*np.sign(self.direction[1])
            
            dc = -(self.arrowL)*self.direction + np.array([dx, dy])
            dc -= np.array([2, 1.5])
            self.__tip_to_center = np.linalg.norm(dc)
            
            deltaCenter = self.tip_pos + qtc.QPointF(*dc)
            
            self.framePos = deltaCenter - qtc.QPointF(w/2, h/2)

    def paintEvent(self, event):
        painter = qtg.QPainter()
        
        painter.begin(self)
        
        painter.setRenderHint(qtg.QPainter.Antialiasing)
        painter.setFont(self.font())
        painter.setPen(self.paintPen)
        
        self.paintArrow(painter)
        self.paintTextBox(painter)
        
        painter.end()

    def paintArrow(self, painter):
        
        painter.save()
        painter.translate(self.tip_pos)
        painter.rotate(self.angle)
        
        painter.drawPolyline(qtc.QPointF(-self.headL, -self.headW/2),
                             qtc.QPointF(0, 0),
                             qtc.QPointF(-self.headL, self.headW/2))
        painter.drawLine(qtc.QPointF(0, 0), qtc.QPointF(-self.arrowL, 0))
        painter.restore()
    
    def paintTextBox(self, painter):
        painter.translate(self.framePos)
        
        painter.fillPath(self.frame, self.bgColor)
        painter.drawPath(self.frame)
        painter.drawText(self.textBbox, self.text)
    
    def update_position(self):
        """
        Moves the annotation such that self.tip_pos is at the same global
        position as self.head_pos. This function should be called before showing
        the annotation to make sure its position on the screen is correct
        """
        tip = self.tip_pos  # local position
        tip_global = self.mapToGlobal(qtc.QPoint(tip.x(), tip.y()))
        new_pos = self.pos()
        delta = qtc.QPoint(*self.head_pos) - tip_global
        new_pos += delta
        self.move(new_pos)




"""ViPErLEED utility: Rescale cell.

This utility takes a slab in POSCAR format and rescales the unit cell.
"""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-08-03'
__license__ = 'GPLv3+'

from viperleed.cli_base import length_choices
from viperleed.utilities.poscar.base import _PoscarStreamCLI


class RescaleCellCLI(_PoscarStreamCLI, cli_name='rescale_cell'):
    """Apply scaling factors to the unit cell of a POSCAR."""

    long_name = 'rescale unit cell'

    def add_parser_arguments(self, parser):
        """Add mandatory scaling factor(s) argument."""
        super().add_parser_arguments(parser)
        parser.add_argument(
            'scaling',
            help=('One or three scaling factors for the unit cell. If three '
                  'values are given, the scaling factors are applied to the '
                  'a, b, and c vectors, respectively. If only one value is '
                  'given, an isotropic scaling is applied.'),
            type=float,
            nargs='+',
            action=length_choices(1, 3),
            )

    def process_slab(self, slab, args):
        """Return a slab with a unit cell scaled according to args."""
        slab.apply_scaling(*args.scaling)
        return slab


if __name__ == '__main__':
    RescaleCellCLI.run_as_script()



"""ViPErLEED utility: Attach Bulk

Takes a slab POSCAR and adds a bulk POSCAR on the bottom, rescaling the
unit cell. Very primitive script, should be updated to include more
recent functionality.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2019-06-14'
__license__ = 'GPLv3+'

import copy
import logging
import time
from timeit import default_timer as timer

import numpy as np

from viperleed.calc.files import poscar
from viperleed.calc.lib.base import get_elapsed_time_str
from viperleed.cli_base import ViPErLEEDCLI


_LOG_FORMAT = '%(levelname)s - %(message)s'
_LOG_HEAD = '''\
Starting new log: {logname}
Time of execution (UTC): {now}
'''
_LOG_TAIL = '''\
Finishing {logname} at {now}.
Elapsed time: {delta_t}
'''
_CONSOLE_HANDLER = logging.StreamHandler()
_CONSOLE_HANDLER.setFormatter(logging.Formatter(_LOG_FORMAT))


class AttachBulkCLI(ViPErLEEDCLI, cli_name='attach_bulk'):
    """Main command-line interface of this utility."""

    def __call__(self, _=None):
        """Call this utility."""
        try:
            return self._call_impl()
        except KeyboardInterrupt:
            return 1
        finally:
            _tear_down_logger()

    @staticmethod
    def _call_impl():
        """The actual implementation of the __call__ method."""
        logname = 'Combine-POSCAR.log'
        _set_up_logger(logname)
        starttime = timer()

        # pylint: disable-next=logging-format-interpolation
        logging.info(_LOG_HEAD.format(logname=logname, now=_now()))
        try:
            slab, bulk = _read_both_poscars_from_user_input()
        except poscar.POSCARError:
            # Error reading. Already reported
            return 2

        try:  # If a and b vectors differ, cancel operation
            _check_ab_consistent(slab, bulk)
        except ValueError as exc:
            logging.error(f'{exc} Stopping operation...')
            return 2

        _attach_bulk(slab, bulk)

        try:
            poscar.write(slab)
        except OSError:
            logging.error('Exception while writing combined POSCAR:',
                          exc_info=True)
        # pylint: disable-next=logging-format-interpolation
        logging.info(_LOG_TAIL.format(
            logname=logname,
            now=_now(),
            delta_t=get_elapsed_time_str(timer() - starttime)
            ))
        return 0


def _add_missing_elements(slab, bulk):
    """Add to slab the extra elements of bulk."""
    if slab.elements == bulk.elements:
        logging.debug('Slab and bulk elements are identical.')
        return
    logging.debug('Slab and bulk elements are not '
                  'equal. Adding missing elements.')
    for element in bulk.elements:
        slab.n_per_elem[element] = 0


def _attach_bulk(slab, bulk):
    """Attach bulk at the bottom of slab."""
    _add_missing_elements(slab, bulk)

    # Resize the slab unit cell
    cfact = slab.c_vector[2]/bulk.c_vector[2]
    slab.c_vector[:] *= (cfact + 1) / cfact
    # Recalculate c for the slab atoms (undistort & shift)
    for atom in slab:
        atom.pos[2] = (atom.pos[2]*(cfact/(cfact+1)))+(1/(cfact+1))
        atom.num += bulk.n_atoms
    # Copy atoms from bulk and add them to the slab
    for atom in bulk:
        newat = copy.copy(atom)
        # Recalculate bulk atom c in the new slab unit cell
        newat.pos[2] /= cfact+1
        slab.atlist.append(newat)
        newat.slab = slab
    slab.atlist.update_atoms_map()
    slab.update_element_count()
    slab.sort_by_element()


def _check_ab_consistent(slab, bulk):
    """Raise if a&b are not identical within 1e-4."""
    eps = 1e-4
    ab_diff = slab.ab_cell.T - bulk.ab_cell.T
    if any(np.linalg.norm(ab_diff, axis=1) > eps):
        raise ValueError('Slab and bulk unit cell vectors are not '
                         f'equal in a and b (error > {eps:g}).')
    logging.debug('Slab and bulk unit cell vectors are '
                  f'equal in a and b (error < {eps:g}).')


def _now():
    """Return the current time as a string."""
    return time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime())


def _read_both_poscars_from_user_input():
    """Return slab and bulk POSCARs to be merged."""
    slab = _read_poscar_from_user_input('slab')
    bulk = _read_poscar_from_user_input('bulk')
    return slab, bulk


def _read_poscar_from_user_input(name):
    """Return a slab read from a file specified by the user."""
    while True:
        filename = input(f'Enter {name} POSCAR name (Ctrl+C to abort): ')
        if not filename:
            print('Input failed. Please try again.')
            continue
        try:
            slab = poscar.read(filename)
        except FileNotFoundError:
            logging.error(f'{filename} not found.')
        except poscar.POSCARError:
            logging.error('Exception while reading POSCAR', exc_info=True)
            raise
        else:
            logging.info(f'{name.capitalize()} POSCAR was read successfully.')
            return slab


def _set_up_logger(logname):
    """Prepare the logging module to log to logname and the console."""
    logging.basicConfig(level=logging.DEBUG, filename=logname,
                        filemode='w', format=_LOG_FORMAT)
    logging.getLogger().addHandler(_CONSOLE_HANDLER)


def _tear_down_logger():
    """Remove handlers and shutdown the logger."""
    logging.getLogger().removeHandler(_CONSOLE_HANDLER)
    logging.shutdown()


if __name__ == '__main__':
    AttachBulkCLI.run_as_script()
